import os
import numpy as np

from PyQt5.QtWidgets import (
    QCheckBox,
    QMessageBox,
    QTableView,
    QComboBox,
    QFileDialog,
    QMainWindow,
    QLabel,
    QLineEdit,
    QPushButton,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QFormLayout,
    QGridLayout,
    QTabWidget,
)

from PyQt5.QtGui import QDoubleValidator, QRegExpValidator
from PyQt5.QtCore import Qt, QSettings, QRegExp
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar

import pandas as pd

from pyboat.ui.util import (
    PandasModel,
    posfloatV,
    posintV,
    default_par_dict,
    selectFilter,
)
from pyboat.ui.analysis import mkTimeSeriesCanvas, FourierAnalyzer, WaveletAnalyzer
from pyboat.ui.batch_process import BatchProcessWindow

import pyboat
from pyboat import plotting as pl

# --- monkey patch label sizes to better fit the ui ---
pl.tick_label_size = 12
pl.label_size = 14

# same for all FileDialogs
FormatFilter = "csv ( *.csv);; MS Excel (*.xlsx);; Text File (*.txt)"


class DataViewer(QMainWindow):
    def __init__(self, data, pos_offset, parent, debug=False):
        super().__init__(parent=parent)

        # this is the data table
        self.df = data

        self.anaWindows = {}  # allows for multiple open analysis windows
        self.w_position = 0  # analysis window position offset

        self.debug = debug

        # this variable tracks the selected trajectory
        # -> DataFrame column name!
        self.signal_id = None  # no signal id initially selected

        self.raw_signal = None  # no signal initial array
        self.dt = None  # gets initialized from the UI -> qset_dt
        self.tvec = None  # gets initialized by vector_prep
        self.time_unit = None  # gets initialized by qset_time_unit

        # get updated with dt in -> qset_dt
        self.periodV = QDoubleValidator(bottom=1e-16, top=1e16, decimals=2)
        self.envelopeV = QDoubleValidator(bottom=3, top=9999999)

        self.initUI(pos_offset)

    # ===========    UI    ================================

    def initUI(self, pos_offset):

        self.setWindowTitle(f"DataViewer - {self.df.name}")
        self.setGeometry(80 + pos_offset, 300 + pos_offset, 900, 650)

        # for the status bar
        main_widget = QWidget()
        self.statusBar()

        self.tsCanvas = mkTimeSeriesCanvas()
        main_frame = QWidget()
        self.tsCanvas.setParent(main_frame)
        ntb = NavigationToolbar(self.tsCanvas, main_frame)  # full toolbar

        # the table instance,
        DataTable = QTableView()
        model = PandasModel(self.df)
        DataTable.setModel(model)
        DataTable.setSelectionBehavior(2)  # columns only
        DataTable.clicked.connect(self.Table_select)  # magically transports QModelIndex
        # so that it also works for header selection
        header = DataTable.horizontalHeader()  # returns QHeaderView
        header.sectionClicked.connect(
            self.Header_select
        )  # magically transports QModelIndex

        # size policy for DataTable, not needed..?!
        # size_pol= QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # DataTable.setSizePolicy(size_pol)

        # the signal selection box
        SignalBox = QComboBox(self)
        SignalBox.setStatusTip("..or just click directly on a signal in the table!")

        main_layout_v = QVBoxLayout()  # The whole Layout
        # Data selction drop-down
        dataLabel = QLabel("Select Signal", self)

        dt_label = QLabel("Sampling Interval:")
        self.dt_edit = QLineEdit()
        self.dt_edit.setStatusTip("How much time in between two recordings?")
        self.dt_edit.setMinimumSize(70, 0)  # no effect :/
        self.dt_edit.setValidator(posfloatV)

        unit_label = QLabel("Time Unit:")
        self.unit_edit = QLineEdit(self)
        self.unit_edit.setStatusTip("Changes only the axis labels..")
        self.unit_edit.setMinimumSize(70, 0)

        top_bar_box = QWidget()
        top_bar_layout = QHBoxLayout()

        top_bar_layout.addWidget(dataLabel)
        top_bar_layout.addWidget(SignalBox)
        top_bar_layout.addStretch(0)
        top_bar_layout.addWidget(dt_label)
        top_bar_layout.addWidget(self.dt_edit)
        top_bar_layout.addStretch(0)
        top_bar_layout.addWidget(unit_label)
        top_bar_layout.addWidget(self.unit_edit)
        top_bar_layout.addStretch(0)
        top_bar_box.setLayout(top_bar_layout)

        top_and_table = QGroupBox("Settings and Data")
        top_and_table_layout = QVBoxLayout()
        top_and_table_layout.addWidget(top_bar_box)
        top_and_table_layout.addWidget(DataTable)
        top_and_table.setLayout(top_and_table_layout)
        main_layout_v.addWidget(top_and_table)

        ## detrending parameter

        self.T_c_edit = QLineEdit()
        self.T_c_edit.setMaximumWidth(70)
        self.T_c_edit.setValidator(posfloatV)
        self.T_c_edit.setStatusTip("..in time units, e.g. 120min")

        sinc_options_box = QGroupBox("Sinc Detrending")
        sinc_options_layout = QGridLayout()
        sinc_options_layout.addWidget(QLabel("Cut-off Period:"), 0, 0)
        sinc_options_layout.addWidget(self.T_c_edit, 0, 1)
        sinc_options_box.setLayout(sinc_options_layout)

        ## Amplitude envelope parameter
        self.wsize_edit = QLineEdit()
        self.wsize_edit.setMaximumWidth(70)
        self.wsize_edit.setValidator(self.envelopeV)
        self.wsize_edit.setStatusTip("..in time units, e.g. 120min")

        envelope_options_box = QGroupBox("Amplitude Envelope")
        envelope_options_layout = QGridLayout()
        envelope_options_layout.addWidget(QLabel("Window Size:"), 0, 0)
        envelope_options_layout.addWidget(self.wsize_edit, 0, 1)
        envelope_options_box.setLayout(envelope_options_layout)

        # plot options box
        plot_options_box = QGroupBox("Plotting Options")
        plot_options_layout = QGridLayout()

        self.cb_raw = QCheckBox("Raw Signal", self)
        self.cb_raw.setStatusTip("Plots the raw unfiltered signal")

        self.cb_trend = QCheckBox("Trend", self)
        self.cb_trend.setStatusTip("Plots the sinc filtered signal")

        self.cb_detrend = QCheckBox("Detrended Signal", self)
        self.cb_detrend.setStatusTip(
            "Plots the signal after trend subtraction (detrending)"
        )

        self.cb_envelope = QCheckBox("Envelope", self)
        self.cb_envelope.setStatusTip("Plots the estimated amplitude envelope")

        plotButton = QPushButton("Refresh Plot", self)
        plotButton.setStatusTip("Updates the plot with the new filter parameter values")
        plotButton.clicked.connect(self.doPlot)

        saveButton = QPushButton("Save Filter Results", self)
        saveButton.clicked.connect(self.save_out_trend)
        saveButton.setStatusTip("Writes the trend and the detrended signal into a file")

        ## checkbox layout
        plot_options_layout.addWidget(self.cb_raw, 0, 0)
        plot_options_layout.addWidget(self.cb_trend, 0, 1)
        plot_options_layout.addWidget(self.cb_detrend, 1, 0)
        plot_options_layout.addWidget(self.cb_envelope, 1, 1)
        plot_options_layout.addWidget(plotButton, 2, 0)
        plot_options_layout.addWidget(saveButton, 2, 1, 1, 1)
        plot_options_box.setLayout(plot_options_layout)

        ## checkbox signal set and change
        self.cb_raw.toggle()

        self.cb_raw.stateChanged.connect(self.toggle_raw)
        self.cb_trend.stateChanged.connect(self.toggle_trend)
        self.cb_detrend.stateChanged.connect(self.toggle_trend)
        self.cb_envelope.stateChanged.connect(self.toggle_envelope)

        # Ploting box/Canvas area
        plot_box = QGroupBox("Signal and Trend")
        plot_layout = QVBoxLayout()
        plot_layout.addWidget(self.tsCanvas)
        plot_layout.addWidget(ntb)
        plot_box.setLayout(plot_layout)

        # Analyzer box with tabs
        ana_widget = QGroupBox("Analysis")
        ana_box = QVBoxLayout()

        ## Initialize tab scresen
        tabs = QTabWidget()
        tab1 = QWidget()
        tab2 = QWidget()

        ## Add tabs
        tabs.addTab(tab1, "Wavelet Analysis")
        tabs.addTab(tab2, "Fourier Transform")

        ## Create first tab
        tab1.parameter_box = QFormLayout()

        ## for wavlet params, button, etc.
        self.Tmin_edit = QLineEdit()
        self.Tmin_edit.setStatusTip("This is the lower period limit")
        self.nT_edit = QLineEdit()
        self.nT_edit.setValidator(QRegExpValidator(QRegExp("[0-9]+")))
        self.nT_edit.setStatusTip("Increase this number for more spectral resolution")
        self.Tmax_edit = QLineEdit()
        self.Tmax_edit.setStatusTip("This is the upper period limit")

        self.pow_max_edit = QLineEdit()
        self.pow_max_edit.setStatusTip(
            "Enter a fixed value or leave blank for automatic wavelet power scaling"
        )

        Tmin_lab = QLabel("Lowest period")
        step_lab = QLabel("Number of periods")
        Tmax_lab = QLabel("Highest  period")
        pow_max_lab = QLabel("Expected maximal power")

        Tmin_lab.setWordWrap(True)
        step_lab.setWordWrap(True)
        Tmax_lab.setWordWrap(True)
        pow_max_lab.setWordWrap(True)

        wletButton = QPushButton("Analyze Signal", self)
        wletButton.setStyleSheet("background-color: lightblue")
        wletButton.setStatusTip("Opens the wavelet analysis..")
        wletButton.clicked.connect(self.run_wavelet_ana)

        batchButton = QPushButton("Analyze All..", self)
        batchButton.clicked.connect(self.run_batch)
        batchButton.setStatusTip("Starts the batch processing")

        ## add  button to layout
        wlet_button_layout_h = QHBoxLayout()

        wlet_button_layout_h.addStretch(0)
        wlet_button_layout_h.addWidget(wletButton)
        wlet_button_layout_h.addStretch(0)
        wlet_button_layout_h.addWidget(batchButton)
        wlet_button_layout_h.addStretch(0)

        self.cb_use_detrended = QCheckBox("Use Detrended Signal", self)
        # self.cb_use_detrended.stateChanged.connect(self.toggle_trend)
        self.cb_use_detrended.setChecked(True)  # detrend by default

        self.cb_use_envelope = QCheckBox("Normalize with Envelope", self)
        self.cb_use_envelope.stateChanged.connect(self.toggle_envelope)
        self.cb_use_envelope.setChecked(False)  # no envelope by default

        ## Add Wavelet analyzer options to tab1.parameter_box layout

        tab1.parameter_box.addRow(Tmin_lab, self.Tmin_edit)
        tab1.parameter_box.addRow(Tmax_lab, self.Tmax_edit)
        tab1.parameter_box.addRow(step_lab, self.nT_edit)
        tab1.parameter_box.addRow(pow_max_lab, self.pow_max_edit)
        tab1.parameter_box.addRow(self.cb_use_detrended)
        tab1.parameter_box.addRow(self.cb_use_envelope)
        tab1.parameter_box.addRow(wlet_button_layout_h)

        tab1.setLayout(tab1.parameter_box)

        # fourier button
        fButton = QPushButton("Analyze Signal", self)
        ## add  button to layout
        f_button_layout_h = QHBoxLayout()
        fButton.clicked.connect(self.run_fourier_ana)
        f_button_layout_h.addStretch(0)
        f_button_layout_h.addWidget(fButton)

        # fourier detrended switch
        self.cb_use_detrended2 = QCheckBox("Use Detrended Signal", self)
        self.cb_use_detrended2.setChecked(True)  # detrend by default

        self.cb_use_envelope2 = QCheckBox("Normalize with Envelope", self)
        self.cb_use_envelope2.setChecked(False)

        # fourier period or frequency view
        self.cb_FourierT = QCheckBox("Show Frequencies", self)
        self.cb_FourierT.setChecked(False)  # show periods per default

        ## Create second tab
        tab2.parameter_box = QFormLayout()
        # tab2.parameter_box.addRow(Tmin_lab,self.Tmin)
        # tab2.parameter_box.addRow(Tmax_lab,self.Tmax)
        tab2.parameter_box.addRow(self.cb_use_detrended2)
        tab2.parameter_box.addRow(self.cb_use_envelope2)
        tab2.parameter_box.addRow(self.cb_FourierT)
        tab2.parameter_box.addRow(f_button_layout_h)
        tab2.setLayout(tab2.parameter_box)

        # Add tabs to Vbox
        ana_box.addWidget(tabs)
        # set layout of ana_widget (will be added to options layout)
        # as ana_box (containing actual layout)
        ana_widget.setLayout(ana_box)

        # Fix size of table_widget containing parameter boxes
        # -> it's all done via column stretches of
        # the GridLayout below
        # size_pol= QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        # ana_widget.setSizePolicy(size_pol)

        # ==========Plot and Options Layout=================================
        # Merge Plotting canvas and options
        plot_and_options = QWidget()
        layout = QGridLayout()
        plot_and_options.setLayout(layout)

        layout.addWidget(plot_box, 0, 0, 4, 1)
        layout.addWidget(sinc_options_box, 0, 5, 1, 1)
        layout.addWidget(envelope_options_box, 0, 6, 1, 1)
        layout.addWidget(plot_options_box, 1, 5, 1, 2)
        layout.addWidget(ana_widget, 2, 5, 2, 2)

        # plotting-canvas column stretch <-> 1st (0th) column
        layout.setColumnStretch(0, 1)  # plot should stretch
        layout.setColumnMinimumWidth(0, 360)  # plot should not get too small

        layout.setColumnStretch(1, 0)  # options shouldn't stretch
        layout.setColumnStretch(2, 0)  # options shouldn't stretch

        # ==========Main Layout=======================================
        main_layout_v.addWidget(plot_and_options)  # is a VBox

        # populate signal selection box
        SignalBox.addItem("")  # empty initial selector

        for col in self.df.columns:
            SignalBox.addItem(col)

        # connect to plotting machinery
        SignalBox.activated[str].connect(self.select_signal_and_Plot)
        # to modify current index by table selections
        self.SignalBox = SignalBox

        # --- connect some parameter fields ---

        self.dt_edit.textChanged[str].connect(self.qset_dt)
        self.unit_edit.textChanged[str].connect(self.qset_time_unit)

        # --- initialize parameter fields from settings ---

        self.load_settings()

        main_widget.setLayout(main_layout_v)
        self.setCentralWidget(main_widget)
        self.show()

        # trigger initial plot?!
        # self.select_signal_and_Plot(self.df.columns[0])

    # when clicked into the table
    def Table_select(self, qm_index):
        # recieves QModelIndex
        col_nr = qm_index.column()
        self.SignalBox.setCurrentIndex(col_nr + 1)
        if self.debug:
            print("table column number clicked:", col_nr)
        signal_id = self.df.columns[col_nr]  # DataFrame column name
        self.select_signal_and_Plot(signal_id)

    # when clicked on the header
    def Header_select(self, index):
        # recieves index
        col_nr = index
        self.SignalBox.setCurrentIndex(col_nr + 1)

        if self.debug:
            print("table column number clicked:", col_nr)

        signal_id = self.df.columns[col_nr]  # DataFrame column name
        self.select_signal_and_Plot(signal_id)

    # the signal to work on, connected to selection box
    def select_signal_and_Plot(self, text):
        self.signal_id = text
        succ = self.vector_prep(self.signal_id)  # fix a raw_signal + time vector
        if not succ:  # error handling done in data_prep
            print("Could not load", self.signal_id)
            return
        self.set_initial_periods()
        self.set_initial_T_c()
        self.doPlot()

    # probably all the toggle state variables are not needed -> read out checkboxes directly
    def toggle_raw(self, state):
        if state == Qt.Checked:
            self.plot_raw = True
            # untoggle the detrended cb
            self.cb_detrend.setChecked(False)
        else:
            self.plot_raw = False

        # signal selected?
        if self.signal_id:
            self.doPlot()

    def toggle_trend(self, state):

        if self.debug:
            print("new state:", self.cb_trend.isChecked())

        # trying to plot the trend
        if state == Qt.Checked:
            T_c = self.get_T_c(self.T_c_edit)
            if not T_c:
                self.cb_trend.setChecked(False)
                self.cb_detrend.setChecked(False)
                self.cb_use_detrended.setChecked(False)
                return

            # don't plot raw and detrended together (trend is ok)
            if self.cb_detrend.isChecked():
                self.cb_raw.setChecked(False)

        # signal selected?
        if self.signal_id:
            self.doPlot()

    def toggle_envelope(self, state):

        # signal selected?
        if self.signal_id:
            # trying to plot the envelope
            if state == Qt.Checked:
                window_size = self.get_wsize(self.wsize_edit)
                if not window_size:
                    self.cb_envelope.setChecked(False)
                    self.cb_use_envelope.setChecked(False)

            self.doPlot()

    # connected to unit_edit
    def qset_time_unit(self, text):
        self.time_unit = text  # self.unit_edit.text()
        if self.debug:
            print("time unit changed to:", text)

    # connected to dt_edit
    def qset_dt(self, text):

        """
        Triggers the rewrite of the initial periods and
        cut-off period T_c
        """

        # checking the input is done automatically via .setValidator!
        # check,str_val,_ = posfloatV.validate(t,  0) # pos argument not used
        t = text.replace(",", ".")
        try:
            self.dt = float(t)
            self.set_initial_periods(force=False)
            self.set_initial_T_c(force=False)
            # update  Validators
            self.periodV = QDoubleValidator(bottom=2 * self.dt, top=1e16)
            self.envelopeV = QDoubleValidator(
                bottom=3 * self.dt, top=self.df.shape[0] * self.dt
            )

            # refresh plot if a is signal selected
            if self.signal_id:
                self.doPlot()

        # empty input
        except ValueError:
            if self.debug:
                print("dt ValueError", text)
            pass

        if self.debug:
            print("dt set to:", self.dt)

    def get_T_c(self, T_c_edit):

        """
        Uses self.T_c_edit, argument just for clarity. Checks
        for empty input, this function only gets called when
        a detrending operation is requested. Hence, an empty
        QLineEdit will display a user warning and return nothing..
        """

        # value checking done by validator, accepts also comma '1,1' !
        tc = T_c_edit.text().replace(",", ".")
        try:
            T_c = float(tc)
            if self.debug:
                print("T_c set to:", T_c)
            return T_c

        # empty line edit
        except ValueError:

            msgBox = QMessageBox(parent=self)
            msgBox.setWindowTitle("Missing Parameter")
            msgBox.setText("Detrending parameter not set, specify a cut-off period!")
            msgBox.exec()

            if self.debug:
                print("T_c ValueError", tc)
            return None

    def get_wsize(self, wsize_edit):

        # value checking done by validator, accepts also comma '1,1' !
        window_size = wsize_edit.text().replace(",", ".")
        try:
            window_size = float(window_size)

        # empty line edit
        except ValueError:

            msgBox = QMessageBox(parent=self)
            msgBox.setWindowTitle("Missing Parameter")
            msgBox.setText(
                "Amplitude envelope parameter not set, specify a sliding window size!"
            )

            msgBox.exec()

            if self.debug:
                print("window_size ValueError", window_size)
            return None

        if window_size / self.dt < 4:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Out of Bounds")
            msgBox.setText(
                f"""Minimal sliding window size for envelope estimation is {4*self.dt} {self.time_unit}!"""
            )
            msgBox.exec()

            return None

        if window_size / self.dt > self.df.shape[0]:
            max_window_size = self.df.shape[0] * self.dt

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Out of Bounds")
            msgBox.setText(
                f"Maximal sliding window size for envelope estimation is {max_window_size:.2f} {self.time_unit}!"
            )
            msgBox.exec()

            return None

        if self.debug:
            print("window_size set to:", window_size)

        return window_size

    def set_initial_periods(self, force=False):

        """
        rewrite value if force is True
        """

        if self.debug:
            print("set_initial_periods called")

        # check if a Tmin was already entered
        # or rewrite if enforced
        if not bool(self.Tmin_edit.text()) or force:
            self.Tmin_edit.clear()
            self.Tmin_edit.insert(str(2 * self.dt))  # Nyquist

        if np.any(self.raw_signal):  # check if raw_signal already selected
            # check if a Tmax was already entered
            if not bool(self.Tmax_edit.text()) or force:
                # default is 1/4 the observation time
                self.Tmax_edit.clear()
                Tmax_ini = self.dt * 1 / 4 * len(self.raw_signal)
                if self.dt > 0.1:
                    Tmax_ini = int(Tmax_ini)
                self.Tmax_edit.insert(str(Tmax_ini))

    def set_initial_T_c(self, force=False):
        if self.debug:
            print("set_initial_T_c called")

        if np.any(self.raw_signal):  # check if raw_signal already selected
            # check if a T_c was already entered
            if not bool(self.T_c_edit.text()) or force:
                # default is 1.5 * Tmax -> 3/8 the observation time
                self.T_c_edit.clear()

                T_c_ini = self.dt * 3 / 8 * len(self.raw_signal)
                if self.dt > 0.1:
                    T_c_ini = int(T_c_ini)
                else:
                    T_c_ini = np.round(T_c_ini, 3)

                self.T_c_edit.insert(str(T_c_ini))

    def set_wlet_pars(self):

        """
        Retrieves and checks the set wavelet parameters
        of the 'Analysis' input box reading the following
        QLineEdits:

        self.Tmin_edit
        self.Tmax_edit
        self.nT_edit
        self.pow_max_edit

        Further the checkboxes regarding detrending and amplitude
        normalization are evaluated. And

        self.get_wsize()
        self.get_T_c()

        are called if needed.

        Returns
        -------

        wlet_pars : dictionary holding the retrieved parameters,
                    window_size and T_c are set to None if no amplitude
                    normalization or detrending operation should be done

        """

        wlet_pars = {}

        # -- read all the QLineEdits --

        text = self.Tmin_edit.text()
        text = text.replace(",", ".")
        check, _, _ = self.periodV.validate(text, 0)
        if self.debug:
            print("Min periodValidator output:", check, "value:", text)
        if check == 0:

            msgBox = QMessageBox()
            msgBox.setText("Lowest period out of bounds!")
            msgBox.exec()

            return False

        Tmin = float(text)

        if Tmin < 2 * self.dt:

            Tmin = 2 * self.dt
            self.Tmin_edit.clear()
            self.Tmin_edit.insert(str(Tmin))

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Warning")
            msg = f"Lowest period set to Nyquist limit: {Tmin} {self.time_unit}!"
            msgBox.setText(msg)
            msgBox.exec()

        wlet_pars["Tmin"] = Tmin

        step_num = self.nT_edit.text()
        check, _, _ = posintV.validate(step_num, 0)
        if self.debug:
            print("nT posintValidator:", check, "value:", step_num)
        if check == 0:

            msgBox = QMessageBox()
            msgBox.setText("The Number of periods must be a positive integer!")
            msgBox.exec()
            return False

        wlet_pars["step_num"] = int(step_num)
        if int(step_num) > 1000:

            choice = QMessageBox.question(
                self,
                "Too much periods?: ",
                "High number of periods: Do you want to continue?",
                QMessageBox.Yes | QMessageBox.No,
            )
            if choice == QMessageBox.Yes:
                pass
            else:
                return False

        text = self.Tmax_edit.text()
        Tmax = text.replace(",", ".")
        check, _, _ = self.periodV.validate(Tmax, 0)

        if self.debug:
            print("Max periodValidator output:", check)
            print(f"Max period value: {self.Tmax_edit.text()}")
        if check == 0:

            msgBox = QMessageBox()
            msgBox.setText("Highest periods out of bounds!")
            msgBox.exec()

            return False
        wlet_pars["Tmax"] = float(Tmax)

        text = self.pow_max_edit.text()
        pow_max = text.replace(",", ".")
        check, _, _ = posfloatV.validate(pow_max, 0)  # checks for positive float
        if check == 0:

            msgBox = QMessageBox()
            msgBox.setText("Maximal power must be positive!")
            msgBox.exec()

            return False

        # check for empty string:
        if pow_max:
            wlet_pars["pow_max"] = float(pow_max)
        else:
            wlet_pars["pow_max"] = None

        # -- the checkboxes --

        # detrend for the analysis?
        if self.cb_use_detrended.isChecked():
            T_c = self.get_T_c(self.T_c_edit)
            if T_c is None:
                return False  # abort settings
            wlet_pars["T_c"] = T_c
        else:
            # indicates no detrending requested
            wlet_pars["T_c"] = None

        # amplitude normalization is downstram of detrending!
        if self.cb_use_envelope.isChecked():
            window_size = self.get_wsize(self.wsize_edit)
            if window_size is None:
                return False  # abort settings
            wlet_pars["window_size"] = window_size
        else:
            # indicates no ampl. normalization
            wlet_pars["window_size"] = None

        # success!
        return wlet_pars

    def vector_prep(self, signal_id):
        """
        prepares raw signal vector (NaN removal) and
        corresponding time vector
        """
        if self.debug:
            print("preparing", signal_id)

        # checks for empty signal_id string
        if signal_id:
            raw_signal = self.df[signal_id]

            NaNswitches = np.sum(np.diff(np.isnan(raw_signal)))
            if NaNswitches > 1:
                print(f"Warning, non-contiguous NaN region found in {signal_id}!")

                msgBox = QMessageBox()
                msgBox.setText(
                    """
                    Non contiguous regions of missing values
                    encountered, using linear interpolation.

                    Try 'Import..' from the main menu
                    to interpolate missing values for all signals!
                    """
                )
                msgBox.exec()

                self.raw_signal = pyboat.core.interpolate_NaNs(raw_signal)
            else:
                # remove contiguous (like trailing) NaN region
                self.raw_signal = raw_signal[~np.isnan(raw_signal)]

            self.tvec = np.arange(0, len(self.raw_signal), step=1) * self.dt
            return True  # success

        else:

            msgBox = QMessageBox()
            msgBox.setText("Please select a signal!")
            msgBox.exec()
            return False

    def calc_trend(self):

        """ Uses maximal sinc window size """

        T_c = self.get_T_c(self.T_c_edit)
        if not T_c:
            return
        if self.debug:
            print("Calculating trend with T_c = ", T_c)

        trend = pyboat.sinc_smooth(raw_signal=self.raw_signal,
                                   T_c=T_c,
                                   dt=self.dt)
        return trend

    def calc_envelope(self):

        window_size = self.get_wsize(self.wsize_edit)
        if not window_size:
            return
        if self.debug:
            print("Calculating envelope with window_size = ", window_size)

        # cut of frequency set and detended plot activated?
        if self.cb_detrend.isChecked():

            trend = self.calc_trend()
            if trend is None:
                return

            signal = self.raw_signal - trend
        else:
            signal = self.raw_signal

        envelope = pyboat.sliding_window_amplitude(signal,
                                                   window_size,
                                                   dt=self.dt)

        return envelope

    def doPlot(self):

        """
        Checks the checkboxes for trend and envelope..
        """

        # update raw_signal and tvec
        succ = self.vector_prep(self.signal_id)  # error handling done here

        if not succ:
            return False

        if self.debug:
            print(
                "called Plotting [raw] [trend] [detrended] [envelope]",
                self.cb_raw.isChecked(),
                self.cb_trend.isChecked(),
                self.cb_detrend.isChecked(),
                self.cb_envelope.isChecked(),
            )

        # check if trend is needed
        if self.cb_trend.isChecked() or self.cb_detrend.isChecked():
            trend = self.calc_trend()
            if trend is None:
                return
        else:
            trend = None

        # envelope calculation
        if self.cb_envelope.isChecked():
            envelope = self.calc_envelope()
            if envelope is None:
                return

        else:
            envelope = None

        self.tsCanvas.fig1.clf()

        ax1 = pl.mk_signal_ax(self.time_unit, fig=self.tsCanvas.fig1)
        self.tsCanvas.fig1.add_axes(ax1)

        if self.debug:
            print(
                f"plotting signal and trend with {self.tvec[:10]}, {self.raw_signal[:10]}"
            )

        if self.cb_raw.isChecked():
            pl.draw_signal(ax1, time_vector=self.tvec, signal=self.raw_signal)

        if trend is not None and self.cb_trend.isChecked():
            pl.draw_trend(ax1, time_vector=self.tvec, trend=trend)

        if trend is not None and self.cb_detrend.isChecked():
            ax2 = pl.draw_detrended(
                ax1, time_vector=self.tvec, detrended=self.raw_signal - trend
            )
            ax2.legend(fontsize=pl.tick_label_size, loc="lower left")
        if envelope is not None and not self.cb_detrend.isChecked():
            pl.draw_envelope(ax1, time_vector=self.tvec, envelope=envelope)

        # plot on detrended axis
        if envelope is not None and self.cb_detrend.isChecked():
            pl.draw_envelope(ax2, time_vector=self.tvec, envelope=envelope)
            ax2.legend(fontsize=pl.tick_label_size)

        self.tsCanvas.fig1.subplots_adjust(bottom=0.15, left=0.15, right=0.85)

        # add a simple legend
        ax1.legend(fontsize=pl.tick_label_size)

        self.tsCanvas.draw()
        self.tsCanvas.show()

    def run_wavelet_ana(self):
        """ run the Wavelet Analysis """

        if not np.any(self.raw_signal):

            msgBox = QMessageBox()
            msgBox.setWindowTitle("No Signal")
            msgBox.setText("Please select a signal first!")
            msgBox.exec()

            return False

        wlet_pars = self.set_wlet_pars()  # Error handling done there
        if not wlet_pars:
            if self.debug:
                print("Wavelet parameters could not be set!")
            return False

        if self.cb_use_detrended.isChecked():
            trend = self.calc_trend()
            signal = self.raw_signal - trend
        else:
            signal = self.raw_signal

        if self.cb_use_envelope.isChecked():
            window_size = self.get_wsize(self.wsize_edit)
            signal = pyboat.normalize_with_envelope(signal, window_size, dt=self.dt)

        self.w_position += 20

        self.anaWindows[self.w_position] = WaveletAnalyzer(
            signal=signal,
            dt=self.dt,
            Tmin=wlet_pars["Tmin"],
            Tmax=wlet_pars["Tmax"],
            pow_max=wlet_pars["pow_max"],
            step_num=wlet_pars["step_num"],
            position=self.w_position,
            signal_id=self.signal_id,
            time_unit=self.time_unit,
            DEBUG=self.debug,
            parent=self,
        )

    def run_batch(self):

        """
        Takes the ui wavelet settings and
        spwans the batch processing Widget
        """

        # reads the wavelet analysis settings from the ui input
        wlet_pars = self.set_wlet_pars()  # Error handling done there
        if not wlet_pars:
            return

        if self.debug:
            print(f"Started batch processing with {wlet_pars}")

        # Spawning the batch processing config widget
        # is bound to parent Wavelet Window
        self.bc = BatchProcessWindow(self, self.debug)
        self.bc.initUI(wlet_pars)

        return
        print("batch processing done!")

    def run_fourier_ana(self):
        if not np.any(self.raw_signal):

            msgBox = QMessageBox()
            msgBox.setWindowTitle("No Signal")
            msgBox.setText("Please select a signal first!")
            msgBox.exec()
            return False

        # shift new analyser windows
        self.w_position += 20

        if self.cb_use_detrended2.isChecked():
            trend = self.calc_trend()
            signal = self.raw_signal - trend
        else:
            signal = self.raw_signal

        if self.cb_use_envelope2.isChecked():
            window_size = self.get_wsize(self.wsize_edit)
            signal = pyboat.normalize_with_envelope(signal, window_size, self.dt)

        # periods or frequencies?
        if self.cb_FourierT.isChecked():
            show_T = False
        else:
            show_T = True

        self.anaWindows[self.w_position] = FourierAnalyzer(
            signal=signal,
            dt=self.dt,
            signal_id=self.signal_id,
            position=self.w_position,
            time_unit=self.time_unit,
            show_T=show_T,
            parent=self,
        )

    def save_out_trend(self):

        if not np.any(self.raw_signal):

            msgBox = QMessageBox()
            msgBox.setWindowTitle("No Signal")
            msgBox.setText("Please select a signal first!")
            msgBox.exec()
            return

        if self.debug:
            print("saving trend out")

        # -------calculate trend and detrended signal------------
        trend = self.calc_trend()
        dsignal = self.raw_signal - trend

        # add everything to a pandas data frame
        data = np.array([self.raw_signal, trend, dsignal]).T  # stupid pandas..
        columns = ["raw", "trend", "detrended"]
        df_out = pd.DataFrame(data=data, columns=columns)
        # ------------------------------------------------------

        if self.debug:
            print("df_out", df_out[:10])
            print("trend", trend[:10])
        dialog = QFileDialog()

        settings = QSettings()
        # ----------------------------------------------------------
        base_name = str(self.signal_id).replace(" ", "-")
        dir_path = settings.value("dir_name", os.path.curdir)
        data_format = settings.value("data_format", "csv")
        default_name = os.path.join(dir_path, base_name + "_trend.")
        default_name += data_format
        # -----------------------------------------------------------
        file_name, sel_filter = dialog.getSaveFileName(
            self, "Save as", default_name, FormatFilter, selectFilter[data_format]
        )
        # dialog cancelled
        if not file_name:
            return

        file_ext = file_name.split(".")[-1]

        if self.debug:
            print("selected filter:", sel_filter)
            print("out-path:", file_name)
            print("extracted extension:", file_ext)

        if file_ext not in ["txt", "csv", "xlsx"]:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Unknown File Format")
            msgBox.setText("Please append .txt, .csv or .xlsx to the file name!")
            msgBox.exec()
            return

        # ------the write out calls to pandas----------------

        # defaults to 3 decimals
        float_format = settings.value("float_format", "%.3f")

        if file_ext == "txt":
            df_out.to_csv(file_name, index=False, sep="\t", float_format=float_format)

        elif file_ext == "csv":
            df_out.to_csv(file_name, index=False, sep=",", float_format=float_format)

        elif file_ext == "xlsx":
            df_out.to_excel(file_name, index=False, float_format=float_format)

        else:
            if self.debug:
                print("Something went wrong during save out..")
            return
        if self.debug:
            print("Saved!")

    def load_settings(self):

        settings = QSettings()

        # map parameter keys to edits
        key_to_edit = {
            "dt": self.dt_edit,
            "time_unit": self.unit_edit,
            "cut_off": self.T_c_edit,
            "window_size": self.wsize_edit,
            "Tmin": self.Tmin_edit,
            "Tmax": self.Tmax_edit,
            "nT": self.nT_edit,
            "pow_max": self.pow_max_edit,
            "float_format": None,
            "graphics_format": None,
            "data_format": None,
        }

        # load defaults from dict or restore values
        for key, value in default_par_dict.items():
            val = settings.value(key, value)
            edit = key_to_edit[key]
            # some fields might be left empty for dynamic defaults
            if edit and (val is not None):
                edit.clear()  # to be on the safe side
                edit.insert(str(val))



###########################################################################
# Tools for time-frequency analysis with Morlet Wavelets
# Inspired by 'A Practical Guide to Wavelet Analysis' from Torrence
# and Compo 1998
# and 'Identification of Chirps with Continuous Wavelet Transform'
# from Carmona,Hwang and Torresani 1995
###########################################################################

import numpy as np
from numpy.fft import rfft, rfftfreq
from numpy.random import uniform, randn, randint, choice
from numpy import pi
from scipy.signal import savgol_filter
import pandas as pd

# global variables
# -----------------------------------------------------------
omega0 = 2 * pi  # central frequency of the mother wavelet
chi2_95 = 5.99  # 0.05 confidence interval
chi2_99 = 9.21  # 0.01 confidence interval

# clip Wavelets at 1/peak_fraction envelope
peak_fraction = 1e8
clip_support = True

# max sinc length, increases performance for long signals
M_max = 2000
# -----------------------------------------------------------


def compute_spectrum(signal, dt, periods):

    """

    Computes the Wavelet spectrum for a given *signal*
    for the given *periods*.

    Parameters
    ----------

    signal  : a sequence
              the time-series to be analyzed,
              you might want to detrend beforehand!

    dt      : the sampling interval scaled to desired time units

    periods : the list of periods to compute the Wavelet spectrum for,
              must have same units as dt!


    Returns
    -------

    modulus : 2d ndarray of reals,
              the Wavelet power spectrum normalized by signal variance

    transform : 2d complex ndarray,
           the Wavelet transform with dimensions len(periods) x len(signal)

    """

    if periods[0] < 2 * dt:
        print()
        print(f"Warning, Nyquist limit is {2*dt:.2f}!!")
        print()

    # -- subtract the mean --
    signal = np.array(signal) - np.mean(signal)

    periods = np.array(periods)
    dt = float(dt)
    sfreq = 1 / dt  # the sampling frequency

    # --------------------------------------------
    scales = scales_from_periods(periods, sfreq, omega0)
    # --------------------------------------------

    # mx_per = 4*len(signal)/((omega0+sqrt(2+omega0**2))*sfreq)
    mx_per = dt * len(signal)
    if max(periods) > mx_per:

        print()
        print("Warning: Very large periods chosen!")
        print("Max. period should be <", np.rint(mx_per))
        print("proceeding anyways...")

    Morlet = mk_Morlet(omega0)
    transform = CWT(signal, Morlet, scales)  # complex wavelet transform
    sig2 = np.var(signal)  # white noise has then mean power of one
    modulus = np.abs(transform) ** 2 / sig2  # normalize with variance of signal

    return modulus, transform


def get_maxRidge_ys(modulus):

    """
    Just pick the time-consecutive modulus
    (squared complex wavelet transform)
    maxima as the ridge.

    Parameters
    ----------

    modulus : 2d ndarray of reals,
              the Wavelet power spectrum (periods x time)
              normalized by signal variance

    Returns
    -------

    ridge_y  : the y-coordinates of the ridge

    """

    ridge_y = np.argmax(modulus, axis=0)

    return ridge_y


def eval_ridge(
    ridge_y, transform, signal, periods, tvec, power_thresh=0, smoothing_wsize=None
):

    """

    Given the ridge y-coordinates, evaluates the spectrum along
    the time axis return the readout along the ridge.

    Parameters
    ----------

    ridge_y :  sequence of indices, has the length of the time axis
               of the spectrum, e.i. the y-coordinates of a ridge

    transform :     2d complex ndarray, holds the complex Wavelet transform
               with dimensions len(periods) x len(signal)

    signal :   1d ndarray, the original signal to be analyzed,
               only needed for variance calculation

    periods :  1d ndarray, the (central) periods of the Morlet wavelets
               used for the Wavelet transform

    tvec :     1d sequency holding the time axis of the signal

    power_thresh : float, minimal power a point (t, y) of a ridge must have
                   to be included in the result, this thresholds the ridge

    smoothing_wsize : int, optional, Savitkzy-Golay smoothing of the ridge


    Returns
    -------

    A DataFrame with the following columns:

    time      : the t-values of the ridge, can have gaps if thresholding!
    periods   : the instantaneous periods
    frequencies : the instantaneous frequencies
    phase    : the arg(z) values
    power     : the Wavelet Power normalized to white noise (<P(WN)> = 1)
    amplitude : the estimated amplitudes of the signal
    (z        : the (complex) z-values of the Wavelet along the ridge)
                not attached as redundant

    Additional attributes:

    dt : the sampling interval
    Nt : the length of the original signal
    """

    sigma2 = np.var(signal)
    modulus = np.abs(transform) ** 2 / sigma2  # normalize with variance of signal

    # calculate here to minimize arguments needed
    dt = tvec[1] - tvec[0]

    Nt = modulus.shape[1]  # number of time points

    # for the DataFrame, to keep absolute time references
    index = np.arange(Nt)

    ridge_per = periods[ridge_y]
    ridge_z = transform[ridge_y, np.arange(Nt)]  # picking the right t-y values !

    ridge_power = modulus[ridge_y, np.arange(Nt)]

    inds = (
        ridge_power > power_thresh
    )  # boolean array of positions exceeding power threshold
    sign_per = ridge_per[inds]  # periods which cross the power threshold
    ridge_t = tvec[inds]
    ridge_phi = np.angle(ridge_z)[inds] % (2 * pi)  # map to [0,2pi]
    sign_power = ridge_power[inds]
    sign_amplitudes = power_to_amplitude(sign_per, sign_power, np.sqrt(sigma2), dt)

    # sign_z = ridge_z[inds] # not needed

    if smoothing_wsize is not None:
        Ntt = len(ridge_per)

        # sanitize ridge smoothing window size
        if Ntt < smoothing_wsize:
            # need an odd window size
            smoothing_wsize = Ntt if Ntt % 2 == 1 else Ntt - 1

        # print('inds:', inds)
        # smoothed maximum estimate of the whole ridge..
        sign_per = savgol_filter(ridge_per, smoothing_wsize, polyorder=3)[inds]

    # set truncated index to allow for
    # proper concatenation along time axis
    # of multiple ridge readouts (summary stats..)

    ridge_data = pd.DataFrame(
        columns=["time", "periods", "phase", "amplitude", "power", "frequencies"],
        index=index[inds],
    )

    ridge_data["time"] = ridge_t
    ridge_data["phase"] = ridge_phi
    ridge_data["power"] = sign_power
    ridge_data["amplitude"] = sign_amplitudes
    ridge_data["periods"] = sign_per
    ridge_data["frequencies"] = 1 / sign_per

    # attach additional data for
    # internal use only
    ridge_data.dt = dt
    ridge_data.Nt = Nt

    MaxPowerPer = ridge_per[
        np.nanargmax(ridge_power)
    ]  # period of highest power on ridge

    return ridge_data


# --- confidence intervals ---


def get_significant_regions(modulus, empirical_background, confidence=chi2_95):

    """
    Given an (empirical) background Fourier power spectrum,
    returns a boolean mask indicating areas of significant
    oscillations within the wavelet power spectrum for
    the desired confidence interval.

    Parameters
    ----------

    modulus : 2d ndarray of reals,
              the wavelet power spectrum normalized by signal variance

    empirical_background: 1d sequence, Fourier estimate of the background.
                          Must hold the powers at exactly the periods used for
                          the wavelet analysis!

    confidence : float, the Chi-squared value at the desired confidence level.
                 Defaults to the 95% confidence interval.

    Returns
    -------

    sign_regions : 2d boolean ndarray with the shape of *modulus*,
                   is True for every significant point of the
                   power spectrum

    """

    # every period needs a background power value
    # (constant 1 for white noise for examle)
    if not len(empirical_background) == modulus.shape[0]:
        raise ValueError("Empirical background doesn't fit" " to wavelet spectrum!")

    # remove DOF factor from chi2
    conf = confidence / 2.0

    # rescale every column along the period axis
    # with the assumed 0-model spectrum
    scaled_mod = np.zeros(modulus.shape)
    for i, col in enumerate(modulus.T):
        scaled_mod[:, i] = col / empirical_background

    # return boolean mask of the power spectrum
    # indicating significant oscillations
    sign_regions = scaled_mod > conf

    return sign_regions


# --- COI business ---


def get_COI_branches(time_vector):

    """
    Get the left and right branches of the COI

    Parameters
    ----------

    time_vector : ndarray, the complete time vector along the signal
    """

    coi_slope = Morlet_COI()

    N_2 = int(len(time_vector) / 2.0)

    # ascending left side
    left = coi_slope * time_vector[:N_2]
    left_t = time_vector[:N_2]

    # descending right side
    right = -coi_slope * (time_vector[N_2:] - time_vector[-1])
    right_t = time_vector[N_2:]

    return np.r_[left, right], np.r_[left_t, right_t]


def find_COI_crossing(rd):

    """
    checks for first/last time point
    which is outside the COI on the
    left/right boundary of the spectrum.

    Returns indices!

    Parameters
    ----------

    rd : pandas.DataFrame
        the ridge data from eval_ridge()
    """

    # restore the original complete time vector
    tvec = np.arange(rd.Nt) * rd.dt
    coi, coi_t = get_COI_branches(tvec)

    N2 = len(tvec) // 2

    left_inds = np.intersect1d(np.arange(N2), rd.index)

    # first time point outside left COI
    coi_inds = left_inds[coi[left_inds] > rd.periods[left_inds]]
    # left ridge might be entirely outside COI
    i_left = coi_inds[0] if coi_inds.size > 0 else 0

    right_inds = np.intersect1d(N2 + np.arange(N2), rd.index)

    # last time point outside right COI
    coi_inds = right_inds[coi[right_inds] > rd.periods[right_inds]]
    # right ridge might be entirely outside COI
    i_right = coi_inds[-1] if coi_inds.size > 0 else -1

    return i_left, i_right


# ============ Snake Annealing =====================================


def find_ridge_anneal(landscape, y0, T_ini, Nsteps, mx_jump=2, curve_pen=0):

    """
    Taking an initial straight line guess at *y0* finds a ridge in *landscape* which
    minimizes the cost_func_anneal by the simulated annealing method.

    landscape - time x scales signal representation (modulus of Wavelet transform)
    y0        - initial ridge guess is straight line at scale landscape[y0]
                -> best to set it close to a peak in the Wavelet modulus (*landscape*)
    T_ini     - initial value of the temperature for the annealing method
    Nsteps    - Max. number of steps for the algorithm
    mx_jump   - Max. distance in scale direction covered by the random steps
    curve_pen - Penalty weight for the 2nd derivative of the ridge to estimate ->
                high values lead to  less curvy ridges

    """

    print()
    print("started annealing..")

    incr = np.arange(-mx_jump, mx_jump + 1)  # possible jumps in scale direction
    incr = incr[incr != 0]  # remove middle zero

    Nt = landscape.shape[-1]  # number of time points
    Ns = landscape.shape[0]  # number of scales
    t_inds = np.arange(Nt)
    ys = y0 * np.ones(
        Nt, dtype=int
    )  # initial ridge guess is straight line at scale landscape[y0]

    Nrej = 0

    tfac = 0.01  # still arbitrary :/
    T_ini = T_ini * tfac
    curve_pen = curve_pen * tfac

    T_k = T_ini  # for more natural units ->  0 < T_ini < 100 should be ok

    for k in range(Nsteps):

        F = cost_func_anneal(ys, t_inds, landscape, 0, curve_pen)

        pos = randint(
            0, len(ys), size=1
        )  # choose time position to make random scale jump

        # dealing with the scale domain boundaries
        if ys[pos] >= Ns - mx_jump - 1:
            eps = -1

        elif ys[pos] < mx_jump:
            eps = +1

        # jump!
        else:
            eps = choice(incr, size=1)

        ys[pos] = ys[pos] + eps  # the candidate

        F_c = cost_func_anneal(ys, t_inds, landscape, 0, curve_pen)

        accept = True

        # a locally non-optimal move occured
        if F_c > F:
            u = uniform()

            # reject bad move? exp(-(F_c - F)/T_k) is (Boltzmann) probability for bad move to be accepted
            if u > np.exp(-(F_c - F) / T_k):
                accept = False

        if not accept:
            ys[pos] = ys[pos] - eps  # revert the wiggle
            Nrej += 1

        if accept:
            Nrej = 0

        print(T_k)
        T_k = T_ini / np.log(2 + k)  # update temperature

    print()
    print("annealing done!")
    print("final cost:", F_c)
    print("number of final still steps:", Nrej)
    print("final temperature:", T_k * tfac)
    return ys, F_c


def cost_func_anneal(ys, t_inds, landscape, l, m):

    """
    Evaluates ridge candidate *ys* on *landscape* plus penalizing terms
    for 1st (*l*) and 2nd (*m*) derivative of the ridge curve.
    """

    N = len(ys)
    D = -sum(landscape[ys, t_inds])
    S1 = l * sum(abs(np.diff(ys, 1)))
    S2 = m * sum(abs(np.diff(ys, 2)))

    # print D,S1,S2,D + S1 + S2

    return (D + S1 + S2) / N


# =============== Filters +  Detrending =================================


def smooth(x, window_len=11, window="flat", data=None):
    """
    smooth the data using a window with requested size.

    input:
    x: the input signal
    window_len: the dimension of the smoothing window; should be an odd integer
    window: the type of window from 'flat' or 'extern'

    flat window will produce a moving average smoothing.
    data: if not None, will be used as evaluated window!

    """

    x = np.array(x)

    # use externally derieved window evaluation
    if data is not None:
        window_len = len(data)
        window = "extern"

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        raise ValueError("window must not be shorter than 3")

    if window_len % 2 == 0:
        raise ValueError("window_len should be odd")

    if window not in ["flat", "extern"]:
        raise ValueError("Window is none of 'flat' or 'extern'")

    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-1:-window_len:-1]]
    # print(len(s))
    if window == "flat":  # moving average
        w = np.ones(window_len, "d")

    elif window == "extern":
        w = data

    else:
        w = eval(window + "(window_len)")

    y = np.convolve(w / w.sum(), s, mode="valid")

    return y[int((window_len - 1) / 2) : len(y) - int((window_len - 1) / 2)]


def sinc_filter(M, f_c=0.2):

    """
    Cutoff frequency f_c in sampling frequency unit, max 0.5!
    M is blackman window length and must be even, output length will be M+1.

    """

    # not very effective, but should be gets called only once per convolution

    assert M % 2 == 0, "M must be even!"
    res = []

    for x in np.arange(0, M + 1):

        # this is the non-vectorizable part
        if x == M / 2:
            res.append(2 * pi * f_c)
            continue

        r = np.sin(2 * pi * f_c * (x - M / 2)) / (
            x - M / 2
        )  # the sinc filter unwindowed
        r = r * (
            0.42 - 0.5 * np.cos(2 * pi * x / M) + 0.08 * np.cos(4 * pi * x / M)
        )  # blackman window
        res.append(r)

    res = np.array(res)
    res = res / np.sum(res)

    return res


def sinc_smooth(raw_signal, T_c, dt, M=None):

    """
    Convolve the signal with a sinc filter
    of cut-off period *T_c*.

    Length of the filter controlled by
    M, defaults to length of `raw_signal - 1`
    or `M_max`, depending on which one is smaller

    Maximum `M` is length of `raw_signal - 1`,
    has to be even!
    """

    signal = np.array(raw_signal)
    dt = float(dt)

    # relative cut_off frequency
    f_c = dt / T_c

    if M is None:

        M = len(signal) - 1  # max for sharp roll-off
        # cap default filter length to M_max
        M = M if M < M_max else M_max

        # M needs to be even
        if M % 2 != 0:
            M = M - 1

    w = sinc_filter(M, f_c)  # the evaluated windowed sinc filter
    sinc_smoothed = smooth(signal, data=w)

    return sinc_smoothed


def sliding_window_amplitude(signal, window_size, dt, SGsmooth=True):

    """
    Max - Min sliding window operation
    to estimate amplitude envelope.

    free boundaries -> half the window_size + 1
    at 1st and last entry

    optional:
    Savitzky-Golay smoothing of the
    envelope with the same window size

    Parameters
    ----------

    signal : ndarray, the (detrended) signal
    window_size : int, the window size in time units
    dt : float, the sampling interval
    """

    # get the underlying array
    vector = np.array(signal)

    if window_size > (len(signal) - 1) * dt:
        window_size = (len(signal) - 1) * dt
        print(f"Warning, setting window_size to {window_size}!")

    # window size in sampling interval units
    window_size = int(window_size / dt)

    # has to be odd
    if window_size % 2 != 1:
        window_size = window_size + 1

    # rolling matrix
    # Stack Overflow move
    shape = vector.shape[:-1] + (vector.shape[-1] - window_size + 1, window_size)
    strides = vector.strides + (vector.strides[-1],)
    aa = np.lib.stride_tricks.as_strided(vector, shape=shape, strides=strides)

    # treat the boundaries, fill with NaNs
    b_size = window_size // 2
    b_left = np.zeros((b_size, aa.shape[1]), dtype=float) * np.nan
    b_right = np.zeros((b_size, aa.shape[1]), dtype=float) * np.nan

    for i in range(b_size):
        b_left[i, : b_size + i + 1] = vector[: b_size + 1 + i]
        b_right[-(i + 1), -(b_size + 1 + i) :] = vector[-(b_size + 1 + i) :]

    rm = np.vstack([b_left, aa, b_right])

    # max-min/2 in sliding window
    amplitudes = (np.nanmax(rm, axis=1) - np.nanmin(rm, axis=1)) / 2

    if SGsmooth:
        amplitudes = savgol_filter(amplitudes, window_length=window_size, polyorder=3)

    return amplitudes


def normalize_with_envelope(dsignal, window_size, dt):

    """
    Best to use detrending beforehand!

    Mean subtraction is still always performed.

    Does NOT check for zero-divide errors,
    arrising in cases where the signal is constant.

    Parameters
    ----------

    dsignal : ndarray, the (detrended) signal
    window_size : int, the window size in time units, e.g. 17 minutes
    dt : float, the sampling interval
    """

    # mean subtraction
    signal = dsignal - dsignal.mean()

    if window_size > (len(signal) - 1) * dt:
        window_size = (len(signal) - 1) * dt
        print(f"Warning, setting window_size to {window_size}!")

    # ampl. normalization
    env = sliding_window_amplitude(signal, window_size, dt)
    ANsignal = 1 / env * dsignal

    return ANsignal


# =============WAVELETS=====================================================


def scales_from_periods(periods, sfreq, omega0=2 * pi):
    # conversion from periods to morlet scales
    # strictly admissable version from Torrence-Compo
    scales = (omega0 + np.sqrt(2 + omega0 ** 2)) * periods * sfreq / (4 * pi)
    return scales


def scales_from_periodsNA(periods, sfreq, omega0=2 * pi):
    # conversion from periods to morlet scales
    # fits the non-admissable Morlet definition!
    scales = omega0 * periods * sfreq / (2 * pi)
    return scales


# is normed to have unit energy on all scales! ..to be used with CWT underneath
def mk_Morlet(omega0):
    def Morlet(t, scale):
        res = (
            pi ** (-0.25)
            * np.exp(omega0 * 1j * t / scale)
            * np.exp(-0.5 * (t / scale) ** 2)
        )
        return 1 / np.sqrt(scale) * res

    return Morlet


def gauss_envelope(t, scale):

    """
    The gaussian envelope of the Morlet Wavelet
    """

    return 1 / np.sqrt(scale) * pi ** (-0.25) * np.exp(-0.5 * (t / scale) ** 2)


def inverse_gauss(y, scale):

    """
    Invert the right Gaussian branch to
    determine Morlet support cut off
    """

    a = pi ** (-0.25) * 1 / np.sqrt(scale)
    return scale * np.sqrt(-2 * np.log(y / a))


# allows for complex wavelets, needs scales scaled with sampling freq!
def CWT(signal, wavelet, scales, clip_support=clip_support):

    # test for complexity
    if np.iscomplexobj(wavelet(10, 1)):
        output = np.zeros([len(scales), len(signal)], dtype=complex)
    else:
        output = np.zeros([len(scales), len(signal)])

    # we want to take always the maximum support available
    # .. no we don't -> performance, otherwise convolutions scale with N * N!!
    # vec = np.arange(-len(signal)/2, len(signal)/2) # old default

    for ind, scale in enumerate(scales):

        # Morlet main peak value:
        y0 = gauss_envelope(0, scale)

        if clip_support:
            # support cut off at 1/peak_fraction of Morlet peak
            x_max = int(inverse_gauss(y0 / peak_fraction, scale))
        else:
            x_max = len(signal) / 2

        # print(ind, scale, 2*x_max)

        # max support is length of signal
        if 2 * x_max > len(signal):
            vec = np.arange(-len(signal) / 2, len(signal) / 2)

        else:
            vec = np.arange(-x_max, x_max)

        wavelet_data = wavelet(vec, scale)
        output[ind, :] = np.convolve(signal, wavelet_data, mode="same")
    return output


def Morlet_COI(omega0=omega0):
    # slope of Morlet e-folding time in tau-periods (spectral) view
    m = 4 * pi / (np.sqrt(2) * (omega0 + np.sqrt(2 + omega0 ** 2)))
    return m


def power_to_amplitude(periods, powers, signal_std, dt):

    """
    Rescale Morlet wavelet powers according to the
    definition of

    "On the Analytic Wavelet Transform",
    Jonathan M. Lilly 2010

    to get an amplitude estimate.
    """

    scales = scales_from_periods(periods, 1 / dt)

    kappa = 1 / np.sqrt(scales) * pi ** -0.25
    kappa = kappa * signal_std * np.sqrt(2)

    return np.sqrt(powers) * kappa


# ===== Fourier FFT Spectrum ================================================


def compute_fourier(signal, dt):

    N = len(signal)

    df = 1.0 / (N * dt)  # frequency bins

    rf = rfft(signal, norm="ortho")  # positive frequencies
    # use numpy routine for sampling frequencies
    fft_freqs = rfftfreq(len(signal), d=dt)

    # print(N,dt,df)
    # print(len(fft_freqs),len(rf))

    fpower = np.abs(rf) ** 2 / np.var(signal)

    return fft_freqs, fpower


# ============== AR1 spectrum and simulation =============


def ar1_powerspec(alpha, periods, dt):
    res = (1 - alpha ** 2) / (
        1 + alpha ** 2 - 2 * alpha * np.cos(2 * pi * dt / periods)
    )

    return res


def ar1_sim(alpha, N, sigma=1, x0=None):

    N = int(N)
    sol = np.zeros(N)

    if x0 is None:
        x0 = randn()

    sol[0] = x0

    for i in range(1, N):
        sol[i] = alpha * sol[i - 1] + sigma * randn()

    return sol


# ========= Utility functions ==============


def complex_average(phis, axis=0):

    """
    Vectorial/directional average on complex plane

    Parameters
    ----------
    phis: ndarray, holding phase values in rad
    axis: int, axis along which to perform the averaging

    """

    Z = np.sum(np.e ** (phis * 1j), axis=axis)

    # normalization
    Z = Z / phis.shape[axis]

    # 1st Order parameter - circular std
    R = np.abs(Z)

    # mean phase
    Psi = np.angle(Z)

    return R, Psi


# =============== NaN - Missing Values interpolation ====================


def interpolate_NaNs(y):

    """
    linearly interpolates through NaNs
    in a sequence of scalars
    """

    bool_ar = np.isnan(y)
    indexer = lambda z: np.nonzero(z)[0]

    # the interpolated values
    interp_yp = np.interp(indexer(bool_ar), indexer(~bool_ar), y[~bool_ar])

    yy = y.copy()

    # replace NaNs with interpolated values
    yy[bool_ar] = interp_yp

    return yy



'''
A short demonstration of the provided
ensemble measures
'''
import numpy as np
import pandas as pd
from pyboat import ensemble_measures as em
from pyboat import WAnalyzer, ssg
from pyboat import plotting as pl 

# set up analyzing instance
periods = np.linspace(5, 60, 100)
dt = 1
wAn = WAnalyzer(periods, dt)

# create a bunch of chirp signals
# with diverging period over time
Nsignals = 50 # times 2
Tstart = 30 # initial period
Tmax = 50   # slowest signal
Nt = 500 # number of samples per signal
signals = [
    ssg.create_noisy_chirp(T1 = Tstart, T2 = Tend, Nt = Nt, eps = 1)
    for Tend in np.linspace(Tstart, Tmax, Nsignals)
]

# add the same number of pure noise signals
noisy_ones = [
    ssg.ar1_sim(alpha = 0.5, Nt = Nt)
    for i in range(Nsignals)
]

# signals ids are just column numbers here
signals = pd.DataFrame(signals + noisy_ones).T

# get the the individual ridge readouts
ridge_results = {}

# store the individual time averaged Wavelet spectra
df_fouriers = pd.DataFrame(index = wAn.periods)

for ID in signals:

    wAn.compute_spectrum(signals[ID], do_plot = False)
    rd = wAn.get_maxRidge(smoothing_wsize = 11)
    ridge_results[ID] = rd

    df_fouriers[ID] = wAn.get_averaged_spectrum()

# the time-averaged power distribution,
# index holds the signal ids if given, otherwise just a column index
powers_series = em.average_power_distribution(
    ridge_results.values(),
    signal_ids = ridge_results.keys())

# is bi-modal!
pl.power_distribution(powers_series)

# filter out the pure noise signals with a power threshold
high_power_ids = powers_series[powers_series > 10].index
high_power_ridge_results = [ridge_results[ID] for ID in high_power_ids]

# creates a tuple of 4 DataFrames, one summary statistic over time              
# for period, amplitude, power and phase each
res = em.get_ensemble_dynamics(high_power_ridge_results)

pl.ensemble_dynamics(*res)

# the Fourier power distribution with pure noise signals
ax = pl.Fourier_distribution(df_fouriers)
ax.set_title('Whole ensemble', pad = -40)

# the Fourier power distribution without pure noise signals
ax = pl.Fourier_distribution(df_fouriers[high_power_ids])
ax.set_title('Noise filtered out', pad = -40)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
import os
from PyQt5.QtWidgets import (
    QCheckBox,
    QAction,
    QMainWindow,
    QApplication,
    QLabel,
    QLineEdit,
    QPushButton,
    QMessageBox,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QGridLayout,
    QComboBox,
)
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtCore import QUrl, QSettings, QRegExp
from PyQt5.QtGui import QDoubleValidator, QRegExpValidator


from pyboat.ui import util
from pyboat.ui.data_viewer import DataViewer
from pyboat.ui.synth_gen import SynthSignalGen
from pyboat import __version__

# matplotlib settings
from matplotlib import rc

rc("text", usetex=False)  # better for the UI


doc_url = "https://github.com/tensionhead/pyBOAT/blob/master/README.md"
gitter_url = "https://gitter.im/pyBOATbase/support"


class MainWindow(QMainWindow):
    def __init__(self, debug):
        super().__init__()

        self.debug = debug

        self.nViewers = 0
        self.DataViewers = {}  # no Viewers yet
        self.initUI()

    def initUI(self):

        self.setGeometry(80, 100, 200, 50)
        self.setWindowTitle(f"pyBOAT {__version__}")

        # Actions for the menu - status bar
        main_widget = QWidget()
        # online help in lower left corner
        self.statusBar()

        quitAction = QAction("&Quit", self)
        quitAction.setShortcut("Ctrl+Q")
        quitAction.setStatusTip("Quit pyBOAT")
        quitAction.triggered.connect(self.close_application)

        openFile = QAction("&Load data", self)
        openFile.setShortcut("Ctrl+L")
        openFile.setStatusTip("Load data")
        openFile.triggered.connect(self.Load_and_init_Viewer)

        ImportMenu = QAction("&Import..", self)
        ImportMenu.setShortcut("Ctrl+I")
        ImportMenu.setStatusTip("Set import options")
        ImportMenu.triggered.connect(self.init_import_menu)

        go_to_settings = QAction("&Default Parameters", self)
        go_to_settings.setStatusTip("Change default analysis parameters")
        go_to_settings.triggered.connect(self.open_settings_menu)

        go_to_doc = QAction("&Documentation..", self)
        go_to_doc.triggered.connect(self.open_doc_link)

        go_to_gitter = QAction("&Online support..", self)
        go_to_gitter.triggered.connect(self.open_gitter_link)

        # --- the menu bar ---

        mainMenu = self.menuBar()
        mainMenu.setNativeMenuBar(False)

        fileMenu = mainMenu.addMenu("&File")
        fileMenu.addAction(openFile)
        fileMenu.addAction(ImportMenu)
        fileMenu.addAction(quitAction)

        settingsMenu = mainMenu.addMenu("&Settings")
        settingsMenu.addAction(go_to_settings)

        helpMenu = mainMenu.addMenu("&Help")
        helpMenu.addAction(go_to_doc)
        helpMenu.addAction(go_to_gitter)

        # --- Import Data ---

        load_box = QGroupBox("Import Data")
        load_box_layout = QVBoxLayout()

        openFileButton = QPushButton("Open", self)
        openFileButton.setStatusTip(
            "Load a table directly, assumes a header is present!"
        )
        openFileButton.setStyleSheet("background-color: lightblue")

        openFileButton.clicked.connect(self.Load_and_init_Viewer)
        load_box_layout.addWidget(openFileButton)

        ImportButton = QPushButton("Import..", self)
        ImportButton.setStatusTip("Set import options")
        ImportButton.clicked.connect(self.init_import_menu)
        load_box_layout.addWidget(ImportButton)

        load_box.setLayout(load_box_layout)

        # --- SSG ---

        synsig_box = QGroupBox("Synthetic Signals")
        synsig_box_layout = QVBoxLayout()

        synsigButton = QPushButton("Start Generator", self)
        synsigButton.setStyleSheet("background-color: orange")
        synsigButton.setStatusTip("Start the synthetic signal generator")
        synsigButton.clicked.connect(self.init_synsig_generator)

        # quitButton = QPushButton("Quit", self)
        # quitButton.clicked.connect(self.close_application)
        # quitButton.setMaximumWidth(50)

        synsig_box_layout.addStretch(1)
        synsig_box_layout.addWidget(synsigButton)
        synsig_box_layout.addStretch(1)
        synsig_box.setLayout(synsig_box_layout)

        # --- Main Layout ---

        main_layout = QHBoxLayout()
        main_layout.addWidget(load_box)
        main_layout.addWidget(synsig_box)

        # main_layout.addWidget(quitButton,1,1)

        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)
        self.show()

    def init_synsig_generator(self):

        if self.debug:
            print("function init_synsig_generator called..")

        self.ssg = SynthSignalGen(parent=self, debug=self.debug)

    def close_application(self):

        # no confirmation window
        if self.debug:
            appc = QApplication.instance()
            appc.closeAllWindows()
            return

        choice = QMessageBox.question(
            self,
            "Quitting",
            "Do you want to exit pyBOAT?",
            QMessageBox.Yes | QMessageBox.No,
        )
        if choice == QMessageBox.Yes:
            print("Quitting ...")
            # sys.exit()
            appc = QApplication.instance()
            appc.closeAllWindows()
        else:
            pass

    def Load_and_init_Viewer(self):

        if self.debug:
            print("function Viewer_Ini called")

        # retrieve or initialize directory path
        settings = QSettings()
        dir_path = settings.value("dir_name", os.path.curdir)

        # load a table directly
        df, err_msg, dir_path = util.load_data(dir_path, self.debug)

        if err_msg == "cancelled":
            return
        else:
            # save the last directory
            settings.setValue("dir_name", dir_path)

        if err_msg:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Data Import Error")
            msgBox.setText(err_msg)
            msgBox.setDetailedText(
                """Have a look at the example data directory at\n
                github.com/tensionhead/pyBOAT"""
            )
            msgBox.exec()
            return

        self.nViewers += 1
        # initialize new DataViewer with the loaded data
        self.DataViewers[self.nViewers] = DataViewer(
            data=df, pos_offset=self.nViewers * 20, debug=self.debug, parent=self
        )

    def init_import_menu(self):

        self.imp_menu = ImportMenu(self, debug=self.debug)

    def open_settings_menu(self):

        self.settings_menu = SettingsMenu(self, debug=self.debug)

    def open_doc_link(self):

        QDesktopServices.openUrl(QUrl(doc_url))

    def open_gitter_link(self):

        QDesktopServices.openUrl(QUrl(gitter_url))


class ImportMenu(QWidget):
    def __init__(self, parent, debug=False):
        super().__init__()

        self.parent = parent
        self.debug = debug
        self.DataViewers = {}  # no Viewers yet

        self.initUI()

    def initUI(self):

        self.setWindowTitle("Data Import Options")
        self.setGeometry(120, 150, 150, 50)

        config_w = QWidget()
        config_grid = QGridLayout()
        config_w.setLayout(config_grid)

        self.cb_use_ext = QCheckBox("Separator from extension")
        self.cb_use_ext.toggle()
        self.cb_use_ext.setToolTip(
            "Infer the column separator from the file extension like '.csv'"
        )
        self.cb_use_ext.stateChanged.connect(self.toggle_ext)
        self.sep_label = QLabel("Column separator")
        self.sep_label.setDisabled(True)
        self.sep_edit = QLineEdit()
        self.sep_edit.setDisabled(True)
        self.sep_edit.setToolTip("Leave empty for automatic detection")

        self.cb_header = QCheckBox("No header row present", self)
        self.cb_header.setToolTip("Assigns a numeric sequence as column names")
        self.cb_header.setChecked(False)

        self.cb_NaN = QCheckBox("Interpolate missing values")
        self.cb_NaN.setToolTip("Linear interpolate in between missing values")

        NaN_label = QLabel("Set missing values entry")
        self.NaN_edit = QLineEdit()
        tt = """ 
        The following values are interpreted as missing values per default: 
        ‘’, ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’, 
        ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’, 
        ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
        """
        self.NaN_edit.setToolTip(tt)

        config_grid.addWidget(self.cb_use_ext, 0, 0, 1, 2)
        config_grid.addWidget(self.sep_label, 1, 0)
        config_grid.addWidget(self.sep_edit, 1, 1)
        config_grid.addWidget(self.cb_header, 2, 0, 1, 2)
        config_grid.addWidget(self.cb_NaN, 3, 0, 1, 2)
        config_grid.addWidget(NaN_label, 4, 0)
        config_grid.addWidget(self.NaN_edit, 4, 1)

        OpenButton = QPushButton("Open", self)
        OpenButton.setStyleSheet("background-color: lightblue")
        OpenButton.setToolTip("Select an input file with the chosen settings..")
        OpenButton.clicked.connect(self.import_file)

        button_box = QHBoxLayout()
        button_box.addStretch(1)
        button_box.addWidget(OpenButton)
        button_box.addStretch(1)
        button_w = QWidget()
        button_w.setLayout(button_box)

        main_layout = QVBoxLayout()
        main_layout.addWidget(config_w)
        main_layout.addWidget(button_w)

        # set main layout
        self.setLayout(main_layout)
        self.show()

    def toggle_ext(self):

        self.sep_label.setDisabled(self.cb_use_ext.isChecked())
        self.sep_edit.setDisabled(self.cb_use_ext.isChecked())

    def import_file(self):

        """
        Reads the values from the config grid
        and prepares kwargs for the pandas read_... functions

        NaN interpolation is done here if requested
        """

        kwargs = {}

        if self.cb_header.isChecked():
            header = None  # pandas will assign numeric column names
        else:
            header = "infer"

        if not self.cb_use_ext.isChecked():

            sep = self.sep_edit.text()
            # empty field
            if sep == "":
                sep = None
            kwargs["sep"] = sep
            if self.debug:
                print(f"Separator is {sep}, with type {type(sep)}")

        NaN_value = self.NaN_edit.text()
        # empty field
        if NaN_value == "":
            NaN_value = None
        if self.debug:
            print(f"NaN value is {NaN_value}, with type {type(NaN_value)}")

        # assemble key-words for pandas read_... functions
        kwargs["header"] = header
        kwargs["na_values"] = NaN_value

        if self.debug:
            print(f"kwargs for load_data: {kwargs}")

        # retrieve or initialize directory path
        settings = QSettings()
        dir_path = settings.value("dir_name", os.path.curdir)

        # -----------------------------------------------------
        df, err_msg, dir_path = util.load_data(dir_path, debug=self.debug, **kwargs)

        if err_msg != "cancelled":
            # save the last directory
            settings.setValue("dir_name", dir_path)
        elif err_msg == "cancelled":
            return

        if err_msg:
            msgBox = QMessageBox()
            msgBox.setWindowTitle("Data Import Error")
            msgBox.setText(err_msg)
            msgBox.exec()
            return
        # -----------------------------------------------------

        if self.cb_NaN.isChecked():

            N_NaNs = df.isna().to_numpy().sum()
            if N_NaNs > 0:
                msg = f"Found {N_NaNs} missing values in total\nlinearly interpolating through.."
                msgBox = QMessageBox()
                msgBox.setWindowTitle("NaNs detected")
                msgBox.setText(msg)
                msgBox.exec()

            else:
                msgBox = QMessageBox()
                msgBox.setWindowTitle("NaN Interpolation")
                msgBox.setText("No missing values found!")
                msgBox.exec()

            name = df.name
            df = util.interpol_NaNs(df)
            df.name = name  # restore name

        # initialize new DataViewer with the loaded data
        self.parent.nViewers += 1
        self.parent.DataViewers[self.parent.nViewers] = DataViewer(
            data=df, pos_offset=self.parent.nViewers * 20, debug=self.debug, parent=self
        )

        self.close()


class SettingsMenu(QMainWindow):
    def __init__(self, parent, debug=False):
        super().__init__(parent=parent)

        self.parent = parent
        self.debug = debug
        self.DataViewers = {}  # no Viewers yet

        self.initUI()

    def initUI(self):

        self.setWindowTitle("Change Default Parameters")
        self.setGeometry(150, 180, 500, 50)

        config_w = QWidget()
        config_grid = QGridLayout()
        config_w.setLayout(config_grid)

        dt_label = QLabel("Sampling Interval")
        self.dt_edit = QLineEdit()
        self.dt_edit.setValidator(QDoubleValidator(0, 99999, 3))
        self.dt_edit.setToolTip("How much time in between two recordings?")
        time_unit_label = QLabel("Time Unit")
        self.time_unit_edit = QLineEdit()
        self.time_unit_edit.setToolTip("Sets the time unit label")

        cut_off_label = QLabel("Cut-off Period")
        self.cut_off_edit = QLineEdit()
        self.cut_off_edit.setValidator(QDoubleValidator(0, 99999, 3))
        tt = """
        Larger periods get removed by the sinc filter
        Leave blank for pyBOATs dynamic defaults..
        """
        self.cut_off_edit.setToolTip(tt)

        wsize_label = QLabel("Window Size")
        self.wsize_edit = QLineEdit()
        self.wsize_edit.setValidator(QDoubleValidator(0, 99999, 3))
        tt = """
        For amplitude envelope estimation
        Leave blank for pyBOATs dynamic defaults.."""
        self.wsize_edit.setToolTip(tt)

        Tmin_label = QLabel("Smallest Period")
        self.Tmin_edit = QLineEdit()
        self.Tmin_edit.setValidator(QDoubleValidator(0, 99999, 3))
        self.Tmin_edit.setToolTip("Lower period limit for the Wavelet transform")
        Tmax_label = QLabel("Highest Period")
        self.Tmax_edit = QLineEdit()
        self.Tmax_edit.setValidator(QDoubleValidator(0, 99999, 3))
        self.Tmax_edit.setToolTip("Upper period limit for the Wavelet transform")
        nT_label = QLabel("Number of Periods")
        self.nT_edit = QLineEdit()
        self.nT_edit.setValidator(QRegExpValidator(QRegExp("[0-9]+")))
        self.nT_edit.setToolTip("Spectral resolution on the period axis")

        pow_max_label = QLabel("Maximal Power")
        self.pow_max_edit = QLineEdit()
        self.pow_max_edit.setValidator(QDoubleValidator(0, 99999, 3))
        self.pow_max_edit.setToolTip(
            """
            Scales the colormap of the Wavelet spectra
            Leave blank for pyBOATs dynamic defaults.."""
        )

        # 1st column

        config_grid.addWidget(dt_label, 0, 0)
        config_grid.addWidget(self.dt_edit, 0, 1)

        config_grid.addWidget(time_unit_label, 1, 0)
        config_grid.addWidget(self.time_unit_edit, 1, 1)

        config_grid.addWidget(cut_off_label, 2, 0)
        config_grid.addWidget(self.cut_off_edit, 2, 1)

        config_grid.addWidget(wsize_label, 3, 0)
        config_grid.addWidget(self.wsize_edit, 3, 1)

        # 2nd column

        config_grid.addWidget(Tmin_label, 0, 2)
        config_grid.addWidget(self.Tmin_edit, 0, 3)

        config_grid.addWidget(Tmax_label, 1, 2)
        config_grid.addWidget(self.Tmax_edit, 1, 3)

        config_grid.addWidget(nT_label, 2, 2)
        config_grid.addWidget(self.nT_edit, 2, 3)

        config_grid.addWidget(pow_max_label, 3, 2)
        config_grid.addWidget(self.pow_max_edit, 3, 3)

        CloseButton = QPushButton("Close", self)
        CloseButton.setToolTip("Discards not set changes")
        CloseButton.clicked.connect(self.clicked_close)

        RevertButton = QPushButton("Clear!", self)
        RevertButton.setStyleSheet("background-color: red")
        RevertButton.setToolTip("Revert to dynamic defaults")
        RevertButton.clicked.connect(self.clicked_revert)

        OkButton = QPushButton("Set!", self)
        OkButton.setToolTip("Approves changes")
        OkButton.setStyleSheet("background-color: lightblue")
        OkButton.clicked.connect(self.clicked_set)

        button_box = QHBoxLayout()
        button_box.addWidget(RevertButton)
        button_box.addStretch(1)
        button_box.addWidget(CloseButton)
        button_box.addWidget(OkButton)
        # button_box.addStretch(1)
        button_w = QWidget()
        button_w.setLayout(button_box)

        config_box = QGroupBox("Analysis")
        # no extra status bar
        # config_box.setStatusTip(
        # "Clear numeric boxes to revert to pyBOAT's dynamic defaults")

        config_box_layout = QVBoxLayout()
        config_box_layout.addWidget(config_w)
        config_box.setLayout(config_box_layout)

        # fmt options box

        fmt_label = QLabel("Number Format")
        self.fmt_dropdown = QComboBox()
        self.fmt_dropdown.setToolTip("Use scientific for very large or small numbers")
        self.fmt_dropdown.addItem("Decimal")
        self.fmt_dropdown.addItem("Scientific")

        # self.fmt_dropdown.activated[str].connect(self.fmt_choice)

        graphics_label = QLabel("Graphics")
        self.graphics_dropdown = QComboBox()
        self.graphics_dropdown.setToolTip("Graphics format for the batch processing")
        self.graphics_dropdown.addItem("png")
        self.graphics_dropdown.addItem("pdf")
        self.graphics_dropdown.addItem("svg")
        self.graphics_dropdown.addItem("jpg")

        data_label = QLabel("Data")
        self.data_dropdown = QComboBox()
        self.data_dropdown.setToolTip("File format of the various tabular outputs")
        self.data_dropdown.addItem("csv")
        self.data_dropdown.addItem("xlsx")
        self.data_dropdown.addItem("txt")

        output_box = QGroupBox("Output")
        output_box_layout = QHBoxLayout()
        output_box_layout.addWidget(fmt_label)
        output_box_layout.addWidget(self.fmt_dropdown)
        output_box_layout.addStretch(1)
        output_box_layout.addWidget(graphics_label)
        output_box_layout.addWidget(self.graphics_dropdown)
        output_box_layout.addWidget(data_label)
        output_box_layout.addWidget(self.data_dropdown)

        output_box_layout.addStretch(5)
        output_box.setLayout(output_box_layout)

        # map parameter keys to edits
        self.key_to_edit = {
            "dt": self.dt_edit,
            "time_unit": self.time_unit_edit,
            "cut_off": self.cut_off_edit,
            "window_size": self.wsize_edit,
            "Tmin": self.Tmin_edit,
            "Tmax": self.Tmax_edit,
            "nT": self.nT_edit,
            "pow_max": self.pow_max_edit,
            "float_format": None,
            "graphics_format": None,
            "data_format": None,
        }

        # load parameters
        self.load_settings()

        # set main layout
        main_layout = QVBoxLayout()
        main_layout.addWidget(config_box)
        main_layout.addWidget(output_box)
        main_layout.addWidget(button_w)
        main_frame = QWidget()
        main_frame.setLayout(main_layout)

        self.setCentralWidget(main_frame)
        self.show()

    def clicked_set(self):

        """
        Retrieves all input fields
        and saves to QSettings
        """

        settings = QSettings()
        for key, edit in self.key_to_edit.items():

            # setting key has no edit
            if not edit:
                continue

            if key == "time_unit":
                value = edit.text()  # the only string parameter
                settings.setValue(key, value)
                continue
            if key == "nT":
                value = int(edit.text())  # the only integer parameter
                settings.setValue(key, value)
                continue

            value = util.retrieve_double_edit(edit)
            # None is also fine!
            settings.setValue(key, value)

        # the output settings are strings
        if self.fmt_dropdown.currentText() == "Decimal":
            settings.setValue("float_format", "%.3f")
        elif self.fmt_dropdown.currentText() == "Scientific":
            settings.setValue("float_format", "%e")

        # we can take the items directly
        settings.setValue("graphics_format", self.graphics_dropdown.currentText())

        # we can take the items directly
        settings.setValue("data_format", self.data_dropdown.currentText())

        if self.debug:
            for key in settings.allKeys():
                print(f"Set: {key} -> {settings.value(key)}")

    def clicked_close(self):
        self.close()

    def load_settings(self):

        settings = QSettings()

        # load defaults from dict or restore values
        for key, value in util.default_par_dict.items():
            val = settings.value(key, value)
            edit = self.key_to_edit[key]
            # some fields left empty for dynamic defaults
            if edit and (val is not None):
                edit.clear()
                edit.insert(str(val))
            elif edit and val is None:
                edit.clear()

        # load combo box defaults, only works via setting the index :/
        default = util.default_par_dict["float_format"]
        float_format = settings.value("float_format", default)
        map_to_ind = {default: 0, "%e": 1}
        self.fmt_dropdown.setCurrentIndex(map_to_ind[float_format])

        default = util.default_par_dict["graphics_format"]
        graphics_format = settings.value("graphics_format", default)
        map_to_ind = {default: 0, "pdf": 1, "svg": 2, "jpg": 3}
        self.graphics_dropdown.setCurrentIndex(map_to_ind[graphics_format])

        default = util.default_par_dict["data_format"]
        data_format = settings.value("data_format", default)
        map_to_ind = {default: 0, "xlsx": 1, "txt": 2}
        self.data_dropdown.setCurrentIndex(map_to_ind[data_format])

    def clicked_revert(self):

        settings = QSettings()

        # load defaults from dict
        for key, value in util.default_par_dict.items():
            if self.debug:
                print(f"Set: {key} -> {settings.value(key)}")
            settings.setValue(key, value)

        # to update the display
        self.load_settings()



"""OOP API for Wavelet Analysis of pyBOAT"""

import matplotlib.pyplot as ppl
import numpy as np

import pyboat.core as core
import pyboat.plotting as pl
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# for publications monkey patch font sizes
# pl.label_size = 24
# pl.tick_label_size = 20
# -----------------------------------------------------------


class WAnalyzer:

    """
    Convenience class to access all of pyBOATs capabilities
    through an object oriented API.

    Parameters
    ----------
    periods   : ordered sequence of periods to compute the Wavelet spectrum for,
                must have same units as dt!

    dt        : float
                the sampling interval scaled to desired time units, e.g. use
                dt=120 to get results in seconds or dt=2 to get results in minutes


    p_max      : Maximum power for Wavelet spectrum z-axis colormap range,
                 if *None* scales automatically and individually for each
                 signal. Set to a fixed value to allow for visual comparisons
                 between signals.

    time_unit_label: string
                     the label for the time unit, e.g. 's' or 'min'

    M         : Length of the sinc filter window, defaults to length
                of input signal or `core.M_max`,
                depending on which one is smaller. Set to a lower value to
                speed up sinc-detrending, should be at least around 50
                for a somehwat sharp roll-off.


    Attributes (other than the above given during initialization)
    ----------

    modulus : 2d ndarray
              the real Wavelet power spectrum normalized by signal
              variance, has shape len(periods) x len(signal). Is set to None
              before any analysis is done.

    transform : 2d ndarray
                the complex results of the Wavelet transform with
                shape len(periods) x len(signal). Is set to None
                before any analysis is done.

    ridge_data : Pandas DataFrame
                 Analysis results obtained by ridge extraction and evaluation. See
                 `get_maxRidge` method documentation for details. Is set to None
                 before any analysis is done.

    ax_signal : matplotlib.axes.Axes
                Reference to the signal axis if created with `plot_signal`,
                otherwise is None.

    ax_spec_signal : matplotlib.axes.Axes
                Reference to the signal axis of the spectrum figure
                if created with `compute_spectrum`, otherwise is None.

    ax_spec : matplotlib.axes.Axes
              Reference to the power spectrum axis of the spectrum figure
              if created with `compute_spectrum`, otherwise is None.

    Methods
    -------

    sinc_smooth(signal, T_c)
        Convolves the signal with the sinc filter with
        cut-off period *T_c*. Returns the trend.

    sinc_detrend(signal, T_c)
        Convenience function which returns the detrended
        signal after sinc filtering.

    get_envelope(signal, window_size, SGsmooth = True)
        Uses a sliding window Min-Max operation to estimate
        and return the amplitude envelope.

    normalize_amplitude(signal, window_size)
        Convenience function to estimate
        the amplitude envelope and right away normalizes the signal with
        1/envelope.

    compute_spectrum(signal, T_c=None, window_size=None, do_plot=True, draw_coi=False)
        Performs the Wavelet transform, optional ad-hoc filtering if *T_c* and/or
        *window_size* are set. See also `sinc_detrend` and `normalize_amplitude`.
        If *do_plot* is True, creates the *ax_spec* axis.

    get_averaged_spectrum()
        Returns the time averaged Wavelet spectrum, a (very) good
        Fourier estimate.

    get_maxRidge(power_thresh=0, smoothing_wsize=None)
        Computes and evaluates the ridge as consecutive
        maxima of the modulus.

    get_sign_maxRidge(empirical_background, confidence=5.99, smoothing_wsize=None)
        Computes and evaluates the ridge as consecutive
        maxima of the modulus, thresholded by a significance test
        given a background spectrum. Confidence defaults to 95% interval.

    calc_rhythmicity()
        Returns a rhythmicity score: the time averaged ridge power.

    draw_Ridge()
        Draws the ridge on the Wavelet spectrum.

    plot_readout(draw_coi=False, num=None)
        Creates a summary plot after the complete analysis (spectrum+ridge)
        was performed. Shows instantaneous period, phase, amplitude and power.

    plot_signal(signal, legend=False, num=None)
        Creates the signal axis *ax_signal* and plots the signal.

    plot_trend(trend, legend=False)
        Plots the trend on the signal axis.

    plot_envelope(envelope, legend=False, num=None)
        Plots the amplitude envelope on the signal axis.

    plot_FFT(signal, show_periods=True)
        Shows the Fourier power spectrum.
    """

    def __init__(self, periods, dt, p_max=None, time_unit_label="a.u.", M=None):

        # sanitize periods
        if periods[0] < 2 * dt:
            logger.warning(f"Nyquist limit is {2*dt:.2f} {time_unit_label}!!")
            logger.info(f"Setting lower period limit to {2*dt:.2f}")
            periods[0] = 2 * dt

        self.periods = np.linspace(periods[0], periods[-1], len(periods))
        self.dt = dt
        self.p_max = p_max
        self.M = M
        self.time_unit_label = time_unit_label

        # to allow for detrending + Wavelet analysis
        self.ana_signal = None

        # analysis results
        self.transform = None
        self.modulus = None
        self.ridge_data = None

        # the figure axes
        self.ax_signal = None
        self.ax_spec_signal = None
        self.ax_spec = None

    def compute_spectrum(
        self, raw_signal, T_c=None, window_size=None, do_plot=True, draw_coi=False
    ):

        """
        Computes the Wavelet spectrum for a given *raw_signal*.

        Parameters
        ----------
        signal  : a sequence, the time-series to be analyzed
        T_c : float, optional
              Cut off period for the sinc-filter detrending, all periods
              larger than that one are removed from the signal. If not given,
              no sinc-detending will be done.
        window_size : float, optional
                      Length of the sliding window for amplitude
                      envelope estimation in real time units, e.g. 17 minutes.
                      If not given no amplitude normalization will be done.
        do_plot      : boolean, set to False if no plot is desired,
                       good for batch processing
        draw_coi: boolean, set to True if cone of influence
                           shall be drawn on the wavelet power spectrum

        Returns
        -------
        modulus : 2d ndarray
              the real Wavelet power spectrum normalized by signal
              variance, has shape len(periods) x len(signal). Is set to None
              before any analysis is done.
        transform : 2d ndarray
                the complex results of the Wavelet transform with
                shape len(periods) x len(signal). Is set to None
                before any analysis is done.

        After a successful analysis, the analyzer instance updates

        self.transform
        self.modulus

        with the results. If do_plot was set to True, the attributes

        self.ax_spec_signal
        self.ax_spec

        allow to access the created subplots directly.
        """

        if T_c:
            detrended = self.sinc_detrend(raw_signal, T_c)
            ana_signal = detrended
        else:
            ana_signal = raw_signal

        # only after potential detrending!
        if window_size:
            ana_signal = self.normalize_amplitude(ana_signal, window_size)

        self.ana_signal = ana_signal

        modulus, transform = core.compute_spectrum(ana_signal, self.dt, self.periods)

        if do_plot:

            tvec = np.arange(len(ana_signal)) * self.dt

            axs = pl.mk_signal_modulus_ax(self.time_unit_label)
            pl.plot_signal_modulus(
                axs,
                time_vector=tvec,
                signal=ana_signal,
                modulus=modulus,
                periods=self.periods,
                p_max=self.p_max,
            )

            fig = ppl.gcf()
            fig.tight_layout()
            # some space for a title
            fig.subplots_adjust(top=0.95)
            self.ax_spec_signal = axs[0]
            self.ax_spec = axs[1]

            if draw_coi:
                pl.draw_COI(axs[1], time_vector=tvec)

        self.transform = transform
        self.modulus = modulus

        # return also directly
        return modulus, transform

    def get_averaged_spectrum(self):

        """
        Average Wavelet spectrum over time
        to give a Fourier estimate. A Wavelet spectrum
        has to be computed first.

        Returns
        -------

        mfourier : Fourier spectrum estimate
        """

        if self.transform is None:
            logger.warning("Need to compute a wavelet spectrum first!")
            return

        mfourier = np.sum(self.modulus, axis=1) / self.modulus.shape[1]

        return mfourier

    def get_maxRidge(self, power_thresh=0, smoothing_wsize=None):

        """
        Computes and evaluates the ridge as consecutive
        maxima of the modulus.

        Returns the ridge_data DataFrame, see also `core.eval_ridge`!

        Additionally the analyser instance updates

        self.ridge_data

        with the results.


        Parameters
        ----------
        power_thresh : float, threshold for the ridge.
        smoothing_wsize : int, optional
                          Savitkzy-Golay smoothing window size
                          for ridge smoothing

        Returns
        -------
        A DataFrame with the following columns:

        time      : the t-values of the ridge, can have gaps if thresholding!
        periods   : the instantaneous periods
        frequencies : the instantaneous frequencies
        phase    : the instantaneous phases
        power     : the Wavelet Power normalized to white noise (<P(WN)> = 1)
        amplitude : the estimated amplitudes of the signal

        """

        if self.transform is None:
            logger.warning("Need to compute a wavelet spectrum first!")
            return

        # for easy integration
        modulus = self.modulus

        Nt = modulus.shape[1]  # number of time points
        tvec = np.arange(Nt) * self.dt

        # has to be odd
        if smoothing_wsize and smoothing_wsize % 2 == 0:
            smoothing_wsize = smoothing_wsize + 1

        # ================ridge detection=====================================

        ridge_ys = core.get_maxRidge_ys(modulus)

        rd = core.eval_ridge(
            ridge_ys,
            self.transform,
            self.ana_signal,
            self.periods,
            tvec=tvec,
            power_thresh=power_thresh,
            smoothing_wsize=smoothing_wsize,
        )

        self.ridge_data = rd

        # return also directly
        return rd

    def get_sign_maxRidge(
        self, empirical_background, confidence=core.chi2_95, smoothing_wsize=None
    ):

        """
        Computes and evaluates the ridge as consecutive
        maxima of the modulus, thresholded by a significance test for
        the given background spectrum. Confidence defaults to 95% interval.

        Returns the ridge_data DataFrame, see also `core.eval_ridge`!

        Additionally the analyser instance updates

        self.ridge_data

        with the results.


        Parameters
        ----------
        empirical_background : 1d sequence, Fourier estimate of the background.
                               Must hold the powers
                               at exactly the periods used for
                               the wavelet analysis!

        confidence : float, the Chi-squared value at the desired
                    confidence level. Defaults to the 95% confidence interval.

        smoothing_wsize : int, optional
                          Savitkzy-Golay smoothing window size
                          for ridge smoothing

        Returns
        -------
        A DataFrame with the following columns:

        time      : the t-values of the ridge, can have gaps if thresholding!
        periods   : the instantaneous periods
        frequencies : the instantaneous frequencies
        phase    :  the instantaneous phases
        power     : the Wavelet Power normalized to white noise (<P(WN)> = 1)
        amplitude : the estimated amplitudes of the signal

        """

        if self.transform is None:
            logger.warning("Need to compute a wavelet spectrum first!")
            return

        # for easy integration
        modulus = self.modulus

        Nt = modulus.shape[1]  # number of time points
        tvec = np.arange(Nt) * self.dt

        # has to be odd
        if smoothing_wsize and smoothing_wsize % 2 == 0:
            smoothing_wsize = smoothing_wsize + 1

        # ================ridge detection=====================================

        ridge_ys = core.get_maxRidge_ys(modulus)

        rd = core.eval_ridge(
            ridge_ys,
            self.transform,
            self.ana_signal,
            self.periods,
            tvec=tvec,
            power_thresh=0,  # we need the complete ridge here!
            smoothing_wsize=smoothing_wsize,
        )

        spectrum_bool = core.get_significant_regions(self.modulus, empirical_background)

        # boolean mask for the ridge
        ridge_bool = spectrum_bool[ridge_ys, rd.index]

        # the significant parts of the ridge
        sign_ridge = rd[ridge_bool]
        # attach additional data, total length and sampling interval
        sign_ridge.Nt = rd.Nt
        sign_ridge.dt = rd.dt

        self.ridge_data = sign_ridge

        # return also directly
        return sign_ridge

    def calc_rhythmicity(self):

        """
        Returns a rhythmicity score: the time averaged ridge power.
        The averaging is done with respect to total signal length,
        not ridge length! Therefore, thresholded and short but high
        power ridge segments get penalized.
        """

        if self.ridge_data is None:
            logger.warning("Can't calculate rhythmicity, extract a ridge first!")
            return

        R = self.ridge_data.power.sum() / len(self.ana_signal)

        return R

    def sinc_smooth(self, signal, T_c):
        """

        Convolve the signal with a sinc filter
        of cut-off period *T_c*. Returns
        the smoothed signal representing
        the non-linear trend.

        Parameters
        ----------
        signal : a sequence

        T_c : float, Cut off period for the sinc-filter detrending, all periods
              larger than that one are removed from the signal

        Returns
        -------
        trend : numpy 1d-array
        """
        if self.M is not None and self.M > len(signal) - 1:
            logger.warning(f"Max. sinc filter size `M` is {len(signal)-1} samples!")
            self.M = len(signal) - 1

        if self.M is not None and self.M % 2 != 0:
            self.M -= 1
            logger.info(f"Setting `M` to be even to {self.M}")

        trend = core.sinc_smooth(signal, T_c, self.dt, M=self.M)

        return trend

    def sinc_detrend(self, signal, T_c):

        """
        Convenience function which right away subtracts the
        trend obtained by sinc filtering. See 'sinc_smooth'
        for details.
        """

        trend = self.sinc_smooth(signal, T_c)

        detrended = signal - trend

        return detrended

    def get_envelope(self, signal, window_size, SGsmooth=True):

        """
        Max - Min sliding window operation
        to estimate amplitude envelope.

        Parameters
        ----------
        signal : a sequence
        window_size : Length of the sliding window for amplitude
                      envelope estimation in real time units, e.g. 17 minutes
        SGsmooth : bool, optional Savitzky-Golay smoothing of the
                         envelope with the same window size


        Returns
        -------
        ANsignal : ndarray, the amplitude normalized signal
        """

        envelope = core.sliding_window_amplitude(
            signal, window_size, dt=self.dt, SGsmooth=SGsmooth
        )

        return envelope

    def normalize_amplitude(self, signal, window_size):

        """
        Estimates the amplitude envelope with a sliding window
        and normalizes the signal with 1/envelope.

        Best to do a detrending first!
        The signal mean gets subtracted in any case.

        Note that the *window_size* should be at least
        of the length of the lowest period to be expected,
        otherwise oscillatory components get damped.

        Parameters
        ----------
        signal : a sequence
        window_size : Length of the sliding window for amplitude
                      envelope estimation in real time units, e.g. 17 minutes

        Returns
        -------
        ANsignal : ndarray, the amplitude normalized signal

        See Also
        --------
        get_envelope : returns the envelope itself

        """

        ANsignal = core.normalize_with_envelope(signal, window_size, dt=self.dt)

        return ANsignal

    def draw_Ridge(self, marker_size=1.5):

        if self.ridge_data is None:
            logger.warning("Can't draw ridge, need to do a ridge detection first!")
            return

        if self.transform is None:
            logger.warning("Can't draw ridge, plot the spectrum first!")
            return

        if self.ax_spec is None:
            logger.warning("No spectrum plotted, can't draw ridge!")

        pl.draw_Wavelet_ridge(self.ax_spec, self.ridge_data, marker_size)

    def plot_signal(self, signal, num=None, figsize=(6.5, 4.5), **pkwargs):

        """
        Creates the signal-figure and plots the signal.

        Parameters
        ----------
        signal : a sequence
        num : int
              The number of the figure to be created
        figsize : tuple
                  the size of the signal axis in inches (x,y)
        **pkwargs : keyword arguments for the matlotlib `plot`
                    call, e.g. marker='o', lw=3, color='red'
        """

        # create plotting style dictionary from defaults
        # and potentially overwrite with the arguments of **pkwargs
        style_dic = dict(pl.SIGNAL_STYLE, **pkwargs)

        if num or self.ax_signal is None:
            fig = ppl.figure(num, figsize=figsize)
            self.ax_signal = pl.mk_signal_ax(self.time_unit_label, fig=fig)

        tvec = np.arange(len(signal)) * self.dt
        self.ax_signal.plot(tvec, signal, **style_dic)

        if "label" in style_dic:
            self.ax_signal.legend(fontsize=pl.tick_label_size, ncol=2)
            ymin, ymax = self.ax_signal.get_ylim()
            self.ax_signal.set_ylim((ymin, 1.2 * ymax))

        fig = ppl.gcf()
        # fig.subplots_adjust(bottom=0.18)
        fig.tight_layout()

    def plot_trend(self, trend, **pkwargs):

        """
        Plots the trend into the signal axis.

        Parameters
        ----------
        trend : a sequence
        **pkwargs : keyword arguments for the matlotlib `plot`
                    call, e.g. marker='o', lw=3, color='red'
        """

        if self.ax_signal is None:
            logger.warning("No axis to plot trend, plot a signal first!")
            return

        # create plotting style dictionary from defaults
        # and potentially overwrite with the arguments of **pkwargs
        style_dic = dict(pl.TREND_STYLE, **pkwargs)

        tvec = np.arange(len(trend)) * self.dt
        self.ax_signal.plot(tvec, trend, **style_dic)

        if "label" in style_dic:
            self.ax_signal.legend(fontsize=pl.tick_label_size, ncol=2)
            ymin, ymax = self.ax_signal.get_ylim()
            self.ax_signal.set_ylim((ymin, 1.2 * ymax))

        fig = ppl.gcf()
        fig.subplots_adjust(bottom=0.18)
        fig.tight_layout()

    def plot_envelope(self, envelope, **pkwargs):

        """
        Plot the sliding window amplitude envelope onto the signal.

        Parameters
        ----------
        envelope : a sequence
        **pkwargs : keyword arguments for the matlotlib `plot`
                    call, e.g. marker='o', lw=3, color='red'
        """

        if self.ax_signal is None:
            logger.warning("Can't plot envelope, plot a signal first!")

        # create plotting style dictionary from defaults
        # and potentially overwrite with the arguments of **pkwargs
        style_dic = dict(pl.TREND_STYLE, **pkwargs)

        tvec = np.arange(len(envelope)) * self.dt
        self.ax_signal.plot(tvec, envelope, **style_dic)

        if "label" in style_dic:
            self.ax_signal.legend(fontsize=pl.tick_label_size, ncol=2)
            ymin, ymax = self.ax_signal.get_ylim()
            self.ax_signal.set_ylim((ymin, 1.2 * ymax))

        fig = ppl.gcf()
        fig.subplots_adjust(bottom=0.18)
        fig.tight_layout()

    def plot_readout(self, draw_coi=False, num=None):

        """
        Wraps the readout plot from `pyboat.plotting.plot_readout`.
        Set *num* explicit to control the figure instance created.
        Uses default plotting styles.
        """

        if self.ridge_data is None:
            logger.warning("Need to extract a ridge first!")
            return

        pl.plot_readout(
            self.ridge_data, time_unit=self.time_unit_label, draw_coi=draw_coi
        )

    def draw_confidence_from_bg(
        self, empirical_background, confidence=core.chi2_95, **pkwargs
    ):

        """
        Given an (empirical) background Fourier spectrum,
        draws the contours of the 95% confidence interval
        on the Wavelet power spectrum.

        empirical_background: a sequence, must hold the powers
                              at exactly the periods used for
                              the wavelet analysis (self.periods)!

        confidence : float, the Chi-squared value at the desired
                    confidence level. Defaults to the 95% confidence interval.

        **pkwargs : additional plotting options passed to
                    matplotlib's contour()
        """
        if self.transform is None:
            print("Need to compute a wavelet spectrum first!")
            return

        # every period needs a background power value
        # (constant 1 for white noise for examle)
        if not len(empirical_background) == self.modulus.shape[0]:
            raise ValueError("Empirical background doesn't fit" " to wavelet spectrum!")

        tvec = np.arange(self.transform.shape[1]) * self.dt
        x, y = np.meshgrid(tvec, self.periods)

        # remove DOF factor
        confidence = confidence / 2.0

        scaled_mod = np.zeros(self.modulus.shape)

        # rescale every column along the period axis
        # with the assumed 0-model spectrum
        for i, col in enumerate(self.modulus.T):
            scaled_mod[:, i] = col / empirical_background

        default_style = {"linewidths": 1.4, "colors": "orange", "alpha": 0.8}

        # update with user arguments
        style_dic = dict(default_style, **pkwargs)

        CS = self.ax_spec.contour(x, y, scaled_mod, levels=[confidence], **style_dic)

    def draw_AR1_confidence(self, alpha):

        """
        This is a special function which only makes
        sense if AR(1) is the 0-model, and alpha is known!
        Uses the theoretical AR(1) power spectrum to mark
        significant regions in the spectrum. Plots both the
        95% and 99% confidence intervals.
        """

        if self.transform is None:
            logger.warning("Need to compute a wavelet spectrum first!")
            return

        tvec = np.arange(self.transform.shape[1]) * self.dt
        x, y = np.meshgrid(tvec, self.periods)  # for plotting the wavelet transform

        ar1power = core.ar1_powerspec(alpha, self.periods, self.dt)

        scaled_mod = np.zeros(self.modulus.shape)

        # maybe there is a more clever way
        for i, col in enumerate(self.modulus.T):
            scaled_mod[:, i] = col / ar1power

        CS = self.ax_spec.contour(
            x,
            y,
            scaled_mod,
            levels=[core.chi2_95 / 2.0],
            linewidths=1.7,
            colors="0.95",
            alpha=0.8,
        )

        CS = self.ax_spec.contour(
            x,
            y,
            scaled_mod,
            levels=[core.chi2_99 / 2.0],
            linewidths=1.7,
            colors="orange",
            alpha=0.8,
        )

        # check confidence levels on (long) ar1 realisations !
        # print (len(where(scaled_mod > conf95)[0])/prod(transform.shape))
        # should be ~0.05
        # print (len(where(scaled_mod > conf99)[0])/prod(transform.shape))
        # should be ~0.01

    def plot_FFT(self, signal, show_periods=True):
        fig = ppl.figure(figsize=(5, 2.5))

        ax = pl.mk_Fourier_ax(
            fig, time_unit=self.time_unit_label, show_periods=show_periods
        )

        fft_freqs, fft_power = core.compute_fourier(signal, self.dt)
        logger.info(f"mean fourier power: {np.mean(fft_power):.2f}")
        pl.Fourier_spec(ax, fft_freqs, fft_power, show_periods)
        fig.tight_layout()



""" pyBOAT - A Biological Oscillations Analysis Toolkit """

import sys
import os
import argparse

__version__ = "0.9.9"

# the object oriented API
from .api import WAnalyzer

# the core functions
from .core import sinc_smooth
from .core import sliding_window_amplitude
from .core import normalize_with_envelope
from .core import compute_spectrum
from .core import get_maxRidge_ys
from .core import eval_ridge
from .core import interpolate_NaNs


# --------------
# UI Entry Point
# --------------


def main(argv=None):

    from PyQt5.QtWidgets import QApplication
    from PyQt5.QtGui import QIcon
    from PyQt5.QtCore import QSize

    from pyboat.ui import start_menu

    # --- initialize the Qt App ---

    # args get not parsed inside Qt app
    app = QApplication(sys.argv)

    # add an application icon
    abs_path = os.path.dirname(os.path.realpath(__file__))
    icon_path = os.path.join(abs_path, "logo_circ128x128.png")
    icon = QIcon()
    icon.addFile(icon_path, QSize(128, 128))
    app.setWindowIcon(icon)

    # needed for QSettings
    app.setOrganizationName("tensionhead")
    app.setOrganizationDomain("https://github.com/tensionhead")
    app.setApplicationName("pyBOAT")

    # -- parse command line arguments ---

    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    parser.add_argument("--version", action="version", version="pyBOAT " + __version__)
    args = parser.parse_args(argv)

    debug = args.debug

    if debug:
        print(
            """
            ---------------
            DEBUG enabled!!
            ---------------
            """
        )

        screen = app.primaryScreen()
        print("Screen: %s" % screen.name())
        size = screen.size()
        print("Size: %d x %d" % (size.width(), size.height()))
        rect = screen.availableGeometry()
        print("Available: %d x %d" % (rect.width(), rect.height()))

    # this starts up the Program
    window = start_menu.MainWindow(debug)

    return app.exec()
